"""

Templates module for geopyv.

"""
import logging
import numpy as np

log = logging.getLogger(__name__)


class Template:
    """

    Template base class.

    """

    def __init__(self, size):
        """

        Base class for geopyv subset template.

        Parameters
        ----------
        size : int
            Size of the subset.


        Attributes
        ----------
        shape : `str`
            String describing the shape of the subset.
        dimension : `str`
            String describing the meaning of the size attribute.
        size : `int`
            Size of the subset.
        n_px : `int`
            Number of pixels in the subset template.
        coords : `numpy.ndarray` (Nx, 2)
            2D array of subset template coordinates of type `float`.
        subset_mask : `numpy.ndarray` (Nx, 2)
            2D array of coordinates to mask of type `float`.

        """
        self.size = size
        self._check_size_and_type()
        self.n_px = []
        self.coords = []
        self.shape = "None"
        self.dimension = "None"
        self.subset_mask = np.ones(((2 * self.size) + 1, (2 * self.size) + 1)).astype(
            np.intc
        )

    def _check_size_and_type(self):
        """

        Private method to check if size is a positive integer, and if not convert to
        a positive integer.

        """
        # Check size.
        if self.size < 0:
            self.size = int(np.abs(self.size))
            log.warning(
                "Negative subset size specified. "
                "Converted to an absolute value of {} pixels.".format(self.size)
            )
        # Check type of size.
        if isinstance(self.size, int) is False:
            self.size = int(self.size)
            log.warning(
                "Subset size not specified as an integer. "
                "Converted to an integer of {} pixels.".format(self.size)
            )
        return

    def mask(self, centre, mask):
        """

        Method to mask subset based on binary mask from mesh.

        Parameters
        ----------
        centre : `numpy.ndarray` (x,y)
            Centre of subset.
        mask : `numpy.ndarray` (Nx,Ny)
            Mask to be applied to the mesh.
            Value of 0 indicates pixels to mask in template.

        """

        local = mask[
            int(centre[1]) - self.size : int(centre[1]) + self.size + 1,
            int(centre[0]) - self.size : int(centre[0]) + self.size + 1,
        ]
        local_masked = local * self.subset_mask
        self.coords = np.argwhere(local_masked == 1) - self.size
        self.n_px = np.shape(self.coords)[0]


class Circle(Template):
    """

    Circular subset template class.

    """

    def __init__(self, radius=25):
        """

        Class for circular subset template. Subclassed from Template.

        Parameters
        ----------
        radius : int, optional
            Radius of the subset. Defaults to a value of 25.


        Attributes
        ----------
        shape : `str`
            String describing the shape of the subset.
        dimension : `str`
            String describing the meaning of the size attribute.
        size : `int`
            Radius of the subset.
        n_px : `int`
            Number of pixels in the subset template.
        coords : `numpy.ndarray` (Nx, 2)
            2D array of subset template coordinates of type `float`.
        subset_mask : `numpy.ndarray` (Nx, 2)
            2D array of coordinates to mask of type `float`.

        """
        super().__init__(radius)
        self.shape = "circle"
        self.dimension = "radius"

        # Create template for extracting circular subset information by checking if
        # pixels are within the subset radius.
        x, y = np.meshgrid(
            np.arange(-self.size, self.size + 1, 1),
            np.arange(-self.size, self.size + 1, 1),
        )
        dist = np.sqrt(x**2 + y**2)
        x_s, y_s = np.where(dist <= self.size)

        # Create template coordinates array.
        self.n_px = x_s.shape[0]
        self.coords = np.empty((self.n_px, 2), order="F")
        self.coords[:, 0] = (x_s - self.size).astype(float)
        self.coords[:, 1] = (y_s - self.size).astype(float)

        # Modify subset mask.
        x, y = np.meshgrid(
            np.arange(-self.size, self.size + 1, 1),
            np.arange(-self.size, self.size + 1, 1),
        )
        dist = np.sqrt(x**2 + y**2)
        self.subset_mask[dist > self.size] = 0


class Square(Template):
    """

    Square subset template class.

    """

    def __init__(self, length=25):
        """

        Class for square subset template. Subclassed from Template.

        Parameters
        ----------
        length : int, optional
            Half length of the side of the subset. Defaults to a value of 25.


        Attributes
        ----------
        shape : `str`
            String describing the shape of the subset.
        dimension : `str`
            String describing the meaning of the size attribute.
        size : `int`
            Half length of side of the subset.
        n_px : `int`
            Number of pixels in the subset template.
        coords : `numpy.ndarray` (Nx, 2)
            2D array of subset template coordinates of type `float`.
        subset_mask : `numpy.ndarray` (Nx, 2)
            2D array of coordinates to mask of type `float`.

        """
        super().__init__(length)
        self.shape = "square"
        self.dimension = "length"

        # Create template for square subset.
        x_s, y_s = np.meshgrid(
            np.arange(-self.size, self.size + 1, 1),
            np.arange(-self.size, self.size + 1, 1),
        )

        # Create template coordinates array.
        self.n_px = (2 * self.size + 1) ** 2
        self.coords = np.empty((self.n_px, 2), order="F")
        self.coords[:, 0] = np.ravel(x_s)
        self.coords[:, 1] = np.ravel(y_s)



"""

Plots module for geopyv.

"""
import logging
import sys
import cv2
import matplotlib.pyplot as plt
import matplotlib.tri as tri

from matplotlib.collections import LineCollection
import scipy as sp
import numpy as np
import re
import geopyv as gp
import alphashape

plt.rcParams["axes.prop_cycle"] = plt.cycler(
    "color", plt.cm.tab20(np.linspace(0, 1, 20))
)
log = logging.getLogger(__name__)


def warp_visualiser(template, warp, show, block, save):
    title = "Visualise warp"

    f_coord = np.zeros(2)
    f_coords = template.coords
    g_coords = gp._subset_extensions._g_coords(f_coord, warp, f_coords)

    f_alpha = alphashape.alphashape(f_coords, 1)
    f_x, f_y = f_alpha.exterior.xy
    g_alpha = alphashape.alphashape(g_coords, 1)
    g_x, g_y = g_alpha.exterior.xy

    fig, ax = plt.subplots(num=title)

    ax.fill(f_x, f_y, alpha=0.5, color="b")
    ax.fill(g_x, g_y, alpha=0.5, color="r")

    ax.set_axis_off()
    ax.axis("equal")
    plt.tight_layout()

    # Save
    if save is not None:
        plt.savefig(save, dpi=600)

    # Show or close.
    if show is True:
        plt.show(block=block)
    else:
        plt.close(fig)

    return fig, ax


def inspect_subset_warp(data, mask, show, block, save):
    image = cv2.imread(data["images"]["f_img"], cv2.IMREAD_COLOR)
    image_gs = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    image_gs = cv2.GaussianBlur(image_gs, ksize=(5, 5), sigmaX=1.1, sigmaY=1.1)
    x = data["position"]["x"]
    y = data["position"]["y"]
    template_size = data["template"]["size"]
    title = "Inspect subset: f_coord = ({x},{y}) (px)".format(x=x, y=y)

    f_coord = np.asarray([x, y])
    if data["template"]["shape"] == "circle":
        template = gp.templates.Circle(template_size)
    elif data["template"]["shape"] == "square":
        template = gp.templates.Square(template_size)
    f_img = gp.image.Image(data["images"]["f_img"])
    f_coords = gp._subset_extensions._init_reference(
        f_coord, template.coords, f_img.QCQT
    )[0]
    g_coords = gp._subset_extensions._g_coords(f_coord, data["results"]["p"], f_coords)

    # Crop image and mask.
    x_min = (np.round(min(np.min(f_coords[:, 0]), np.min(g_coords[:, 0])), 0)).astype(
        int
    )
    x_max = (np.round(max(np.max(f_coords[:, 0]), np.max(g_coords[:, 0])), 0)).astype(
        int
    )
    y_min = (np.round(min(np.min(f_coords[:, 1]), np.min(g_coords[:, 1])), 0)).astype(
        int
    )
    y_max = (np.round(max(np.max(f_coords[:, 1]), np.max(g_coords[:, 1])), 0)).astype(
        int
    )

    # image = image_gs.astype(np.float32)[y_min : y_max + 1, x_min : x_max + 1]
    if not isinstance(mask, type(None)):
        if type(mask) == np.ndarray:
            if np.shape(mask) == np.shape(image_gs):
                mask = mask.astype(np.float32)[y_min : y_max + 1, x_min : x_max + 1]
                invert_mask = np.abs(mask - 1) * 255
                image = np.maximum(image, invert_mask)

    f_hull = sp.spatial.ConvexHull(f_coords)
    g_hull = sp.spatial.ConvexHull(g_coords)

    fig, ax = plt.subplots(num=title)
    ax.imshow(
        image,
        cmap="gist_gray",
        interpolation="nearest",
        aspect="equal",
    )
    ax.fill(
        f_coords[f_hull.vertices, 0], f_coords[f_hull.vertices, 1], alpha=0.5, color="b"
    )
    ax.fill(
        g_coords[g_hull.vertices, 0], g_coords[g_hull.vertices, 1], alpha=0.5, color="r"
    )

    ax.text(
        3.0,
        -1.0,
        "warp vector: " + str(data["results"]["p"]),
        horizontalalignment="center",
    )

    ax.set_xlim(x_min - 5, x_max + 5)
    ax.set_ylim(y_max + 5, y_min - 5)

    ax.set_axis_off()
    plt.tight_layout()

    # Save
    if save is not None:
        plt.savefig(save, dpi=600)

    # Show or close.
    if show is True:
        plt.show(block=block)
    else:
        plt.close(fig)

    return fig, ax


def inspect_subset(data, mask, show, block, save):
    """

    Function to show the Subset and associated quality metrics.

    Parameters
    ----------
    data : dict
        Subset data dict.
    mask : `numpy.ndarray`
        Subset mask.
    show : bool
        Control whether the plot is displayed.
    block : bool
        Control whether the plot blocks execution until closed.
    save : str
        Name to use to save plot. Uses default extension of `.png`.


    Returns
    -------
    fig :  matplotlib.pyplot.figure
        Figure object.
    ax : matplotlib.pyplot.axes
        Axes object.


    .. note::
        * The figure and axes objects can be returned allowing standard matplotlib
          functionality to be used to augment the plot generated. See the
          :ref:`plots tutorial <Plots Tutorial>` for guidance.

    .. seealso::
        :meth:`~geopyv.subset.SubsetBase.inspect`

    """
    # Load data.
    image = cv2.imread(data["images"]["f_img"], cv2.IMREAD_COLOR)
    image_gs = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    image_gs = cv2.GaussianBlur(image_gs, ksize=(5, 5), sigmaX=1.1, sigmaY=1.1)
    x = data["position"]["x"]
    y = data["position"]["y"]
    template_size = data["template"]["size"]
    sigma_intensity = data["quality"]["sigma_intensity"]
    SSSIG = data["quality"]["SSSIG"]
    title = "Inspect subset: f_coord = ({x},{y}) (px)".format(x=x, y=y)

    # Crop image and mask.
    x_min = (np.round(x, 0) - template_size).astype(int)
    x_max = (np.round(x, 0) + template_size).astype(int)
    y_min = (np.round(y, 0) - template_size).astype(int)
    y_max = (np.round(y, 0) + template_size).astype(int)
    image = image_gs.astype(np.float32)[y_min : y_max + 1, x_min : x_max + 1]
    if not isinstance(mask, type(None)):
        if type(mask) == np.ndarray:
            if np.shape(mask) == np.shape(image_gs):
                mask = mask.astype(np.float32)[y_min : y_max + 1, x_min : x_max + 1]
                invert_mask = np.abs(mask - 1) * 255
                image = np.maximum(image, invert_mask)

    # If a circular subset, mask pixels outside radius.
    if data["template"]["shape"] == "circle":
        x, y = np.meshgrid(
            np.arange(-template_size, template_size + 1, 1),
            np.arange(-template_size, template_size + 1, 1),
        )
        dist = np.sqrt(x**2 + y**2)
        mask = np.zeros(image.shape)
        mask[dist > template_size] = 255
        image = np.maximum(image, mask)

    # Plot figure.
    fig, ax = plt.subplots(num=title)
    ax.imshow(
        image,
        cmap="gist_gray",
        interpolation="nearest",
        aspect="equal",
        extent=(-0.5, 6.5, -0.5, 5.5),
    )
    size_str = r"Size: {} (px); ".format(template_size)
    sigma_str = r"Quality metrics: $\sigma_s$ = {:.2f} (-);".format(sigma_intensity)
    SSSIG_str = r"SSSIG = {:.2E} (-)".format(SSSIG)
    ax.text(
        3.0,
        -1.0,
        size_str + sigma_str + SSSIG_str,
        horizontalalignment="center",
    )
    ax.set_axis_off()
    plt.tight_layout()

    # Save
    if save is not None:
        plt.savefig(save, dpi=600)

    # Show or close.
    if show is True:
        plt.show(block=block)
    else:
        plt.close(fig)

    return fig, ax


def convergence_subset(data, show, block, save):
    """

    Function to plot Subset convergence.

    Parameters
    ----------
    data : dict
        Subset data dict.
    show : bool
        Control whether the plot is displayed.
    block : bool
        Control whether the plot blocks execution until closed.
    save : str
        Name to use to save plot. Uses default extension of `.png`.


    Returns
    -------
    fig :  matplotlib.pyplot.figure
        Figure object.
    ax : matplotlib.pyplot.axes
        Axes object.


    .. note::
        * The figure and axes objects can be returned allowing standard matplotlib
          functionality to be used to augment the plot generated. See the
          :ref:`plots tutorial <Plots Tutorial>` for guidance.

    .. seealso::
        :meth:`~geopyv.subset.SubsetBase.convergence`

    """
    # Load data.
    history = data["results"]["history"]
    max_iterations = data["settings"]["max_iterations"]
    max_norm = data["settings"]["max_norm"]
    tolerance = data["settings"]["tolerance"]
    x = data["position"]["x"]
    y = data["position"]["y"]

    # Create plot.
    title = "Subset convergence: f_coord = ({x},{y}) (px)".format(x=x, y=y)
    fig, ax = plt.subplots(2, 1, sharex=True, num=title)
    ax[0].semilogy(
        history[0, :],
        history[1, :],
        marker="o",
        clip_on=True,
        label="Convergence",
    )
    ax[0].plot(
        [1, max_iterations],
        [max_norm, max_norm],
        "--r",
        label="Threshold",
    )
    ax[0].set_ylabel(r"$\Delta$ Norm (-)")
    ax[0].set_ylim(max_norm / 1000, max_norm * 1000)
    ax[0].set_yticks(
        [
            max_norm * 1000,
            max_norm * 100,
            max_norm * 10,
            max_norm,
            max_norm / 10,
            max_norm / 100,
            max_norm / 1000,
        ]
    )
    ax[1].plot(
        history[0, :],
        history[2, :],
        marker="o",
        clip_on=True,
        label="Convergence",
    )
    ax[1].plot(
        [1, max_iterations],
        [tolerance, tolerance],
        "--r",
        label="Threshold",
    )
    ax[1].set_ylabel(r"$C_{CC}$ (-)")
    ax[1].set_xlabel("Iteration number (-)")
    ax[1].set_xlim(1, max_iterations)
    ax[1].set_ylim(0.0, 1)
    ax[1].set_yticks(np.linspace(0.0, 1.0, 6))
    ax[1].set_xticks(np.linspace(1, max_iterations, max_iterations))
    ax[0].legend(frameon=False)
    plt.tight_layout()

    # Save
    if save is not None:
        plt.savefig(save, dpi=600)

    # Show or close.
    if show is True:
        plt.show(block=block)
    else:
        plt.close(fig)

    return fig, ax


def convergence_mesh(data, quantity, show, block, save):
    """

    Function to plot Mesh convergence.

    Parameters
    ----------
    data : dict
        Mesh data dict.
    quantity : str
        Quantity to plot. Options are "C_ZNCC", "iterations", or "norm".
    show : bool
        Control whether the plot is displayed.
    block : bool
        Control whether the plot blocks execution until closed.
    save : str
        Name to use to save plot. Uses default extension of `.png`.


    Returns
    -------
    fig :  matplotlib.pyplot.figure
        Figure object.
    ax : matplotlib.pyplot.axes
        Axes object.


    .. note::
        * The figure and axes objects can be returned allowing standard matplotlib
          functionality to be used to augment the plot generated. See the
          :ref:`plots tutorial <Plots Tutorial>` for guidance.

    .. seealso::
        :meth:`~geopyv.mesh.MeshBase.convergence`

    """
    # Get image names.
    try:
        platform = sys.platform
        if platform == "linux" or platform == "linux2" or platform == "darwin":
            split = "/"
        elif platform == "win32":
            split = r"\\"
        f_img = data["images"]["f_img"][
            ([(m.end(0)) for m in re.finditer(split, data["images"]["f_img"])][-1]) :
        ]
        g_img = data["images"]["g_img"][
            ([(m.end(0)) for m in re.finditer(split, data["images"]["g_img"])][-1]) :
        ]
        title = "Convergence: f_img = {f_img}; g_img = {g_img}".format(
            f_img=f_img, g_img=g_img
        )
    except Exception:
        log.warning("Could not extract image names for plot.")
        title = "Convergence:"

    # Extract variables from data.
    subsets = data["results"]["subsets"]
    max_norm = data["settings"]["max_norm"]
    tolerance = data["settings"]["tolerance"]
    max_iterations = data["settings"]["max_iterations"]
    iterations = []
    norm = []
    C_ZNCC = []
    for s in subsets:
        iterations.append(s["results"]["history"][0, -1])
        norm.append(s["results"]["history"][1, -1])
        C_ZNCC.append(s["results"]["history"][2, -1])
    iterations = np.asarray(iterations)
    norm = np.asarray(norm)
    C_ZNCC = np.asarray(C_ZNCC)

    # Create plot.
    fig, ax = plt.subplots(num=title)
    if quantity == "C_ZNCC":
        ax.hist(C_ZNCC, bins=50)
        ax.set_xlabel(r"$C_{ZNCC}$ (-)")
        ax.set_xlim([tolerance, 1.0])
    elif quantity == "iterations":
        ax.hist(iterations)
        ax.set_xlabel(r"Iterations (-)")
        ax.set_xlim([0, max_iterations])
    elif quantity == "norm":
        ax.hist(norm, bins=50)
        ax.set_xlabel(r"$\Delta Norm$ (-)")
        ax.set_xlim([0, max_norm])
    ax.set_ylabel("Count (-)")
    plt.tight_layout()

    # Save
    if save is not None:
        plt.savefig(save, dpi=600)

    # Show or close.
    if show is True:
        plt.show(block=block)
    else:
        plt.close(fig)

    return fig, ax


def contour_mesh(
    data,
    quantity,
    imshow,
    colorbar,
    ticks,
    mesh,
    alpha,
    levels,
    axis,
    xlim,
    ylim,
    show,
    block,
    save,
):
    """Function to plot contours of mesh data."""

    # Load data.
    nodes = data["nodes"]
    elements = data["elements"]
    subsets = data["results"]["subsets"]

    # Extract variables from data.
    x = []
    y = []
    value = []
    for s in subsets:
        x.append(s["position"]["x"])
        y.append(s["position"]["y"])

        if quantity == "u":
            value.append(float((s["results"]["p"])[0]))
        elif quantity == "v":
            value.append(float((s["results"]["p"])[1]))
        elif quantity == "u_x":
            value.append(float((s["results"]["p"])[2]))
        elif quantity == "v_x":
            value.append(float((s["results"]["p"])[3]))
        elif quantity == "u_y":
            value.append(float((s["results"]["p"])[4]))
        elif quantity == "v_y":
            value.append(float((s["results"]["p"])[5]))
        elif quantity == "R":
            value.append(np.sqrt(s["results"]["u"] ** 2 + s["results"]["v"] ** 2))
        else:
            value.append(s["results"][quantity])

    x = np.asarray(x)
    y = np.asarray(y)
    value = np.asarray(value)

    # Plot setup.
    try:
        platform = sys.platform
        if platform == "linux" or platform == "linux2" or platform == "darwin":
            split = "/"
        elif platform == "win32":
            split = r"\\"
        f_img = data["images"]["f_img"][
            ([(m.end(0)) for m in re.finditer(split, data["images"]["f_img"])][-1]) :
        ]
        g_img = data["images"]["g_img"][
            ([(m.end(0)) for m in re.finditer(split, data["images"]["g_img"])][-1]) :
        ]
        title = (
            "Contour: f_img = {f_img}; g_img = {g_img}; variable = {variable}".format(
                f_img=f_img, g_img=g_img, variable=quantity
            )
        )
    except Exception:
        log.warning("Could not extract image names for plot.")
        title = "Contour: variable = {variable}".format(variable=quantity)
    fig, ax = plt.subplots(num=title)

    # Triangulation.
    mesh_triangulation, x_p, y_p = gp.geometry.utilities.plot_triangulation(
        elements, x, y, data["mesh_order"]
    )

    # Plot mesh.
    if mesh is True:
        for i in range(np.shape(x_p)[0]):
            ax.plot(x_p[i], y_p[i], color="k", alpha=0.25, linewidth=0.5)
    triangulation = tri.Triangulation(nodes[:, 0], nodes[:, 1], mesh_triangulation)

    # Set levels and extend.
    extend = "neither"
    if not isinstance(levels, type(None)):
        if np.max(value) > np.max(levels) and np.min(value) < np.min(levels):
            extend = "both"
        elif np.max(value) > np.max(levels):
            extend = "max"
        elif np.min(value) < np.min(levels):
            extend = "min"
    extend = "both"

    # Show image in background.
    if imshow is True:
        image = cv2.imread(data["images"]["f_img"], cv2.IMREAD_COLOR)
        image_gs = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        image_gs = cv2.GaussianBlur(image_gs, ksize=(5, 5), sigmaX=1.1, sigmaY=1.1)
        plt.imshow(image_gs, cmap="gray")
    else:
        ax.set_aspect("equal", "box")

    # Plot contours.
    contours = ax.tricontourf(
        triangulation,
        value,
        alpha=alpha,
        levels=levels,
        extend=extend,
    )
    if colorbar is True:
        if quantity == "iterations":
            label = "Iterations (-)"
        elif quantity == "C_ZNCC":
            label = r"$C_{ZNCC}$ (-)"
        elif quantity == "norm":
            label = r"$\Delta$ Norm (-)"
        else:
            label = quantity
        fig.colorbar(contours, label=label, ticks=ticks)

    # Axis control.
    if axis is False:
        ax.set_axis_off()

    # Limit control.
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    # Save.
    if save is not None:
        plt.savefig(save, dpi=600)

    # Show or close.
    if show is True:
        plt.show(block=block)
    else:
        plt.close(fig)

    return fig, ax


def quiver_mesh(data, scale, imshow, mesh, axis, xlim, ylim, show, block, save):
    """Function to plot quiver plot of mesh data."""

    # Load data.
    elements = data["elements"]
    subsets = data["results"]["subsets"]

    # Extract variables from data.
    x = []
    y = []
    u = []
    v = []
    for s in subsets:
        x.append(s["position"]["x"])
        y.append(s["position"]["y"])
        u.append(s["results"]["u"])
        v.append(s["results"]["v"])
    x = np.asarray(x)
    y = np.asarray(y)
    u = np.asarray(u)
    v = np.asarray(v)
    L = np.sqrt(u**2 + v**2)
    U = u / L
    V = v / L
    m = np.max(L)
    S = scale / (1 + np.log(m / L))
    U1 = S * U * 1 / 2.54
    V1 = S * V * 1 / 2.54

    # Plot setup.
    try:
        platform = sys.platform
        if platform == "linux" or platform == "linux2" or platform == "darwin":
            split = "/"
        elif platform == "win32":
            split = r"\\"
        f_img = data["images"]["f_img"][
            ([(m.end(0)) for m in re.finditer(split, data["images"]["f_img"])][-1]) :
        ]
        g_img = data["images"]["g_img"][
            ([(m.end(0)) for m in re.finditer(split, data["images"]["g_img"])][-1]) :
        ]
        title = "Quiver: f_img = {f_img}; g_img = {g_img}".format(
            f_img=f_img, g_img=g_img
        )
    except Exception:
        log.warning("Could not extract image names for plot.")
        title = "Quiver:"
    fig, ax = plt.subplots(num=title)

    # Plot mesh.
    if mesh is True:
        # Triangulation.
        _, x_p, y_p = gp.geometry.utilities.plot_triangulation(
            elements, x, y, data["mesh_order"]
        )
        for i in range(np.shape(x_p)[0]):
            ax.plot(x_p[i], y_p[i], color="k", alpha=0.25, linewidth=0.5)

    # Show image in background.
    if imshow is True:
        image = cv2.imread(data["images"]["f_img"], cv2.IMREAD_COLOR)
        image_gs = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        image_gs = cv2.GaussianBlur(image_gs, ksize=(5, 5), sigmaX=1.1, sigmaY=1.1)
        plt.imshow(image_gs, cmap="gray")
    else:
        ax.set_aspect("equal", "box")

    # Plot contours.
    ax.quiver(x, y, U1, -V1, color="b", scale=1.0, units="inches")

    # Axis control.
    if axis == "off":
        ax.set_axis_off()

    # Limit control.
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    # Save.
    if save is not None:
        plt.savefig(save, dpi=600)

    # Show or close.
    if show is True:
        plt.show(block=block)
    else:
        plt.close(fig)

    return fig, ax


def inspect_mesh(data, show, block, save):
    """

    Function to inspect Mesh topology.

    Parameters
    ----------
    data : dict
        Mesh data dict.
    show : bool
        Control whether the plot is displayed.
    block : bool
        Control whether the plot blocks execution until closed.
    save : str
        Name to use to save plot. Uses default extension of `.png`.


    Returns
    -------
    fig :  matplotlib.pyplot.figure
        Figure object.
    ax : matplotlib.pyplot.axes
        Axes object.


    .. note::
        * The figure and axes objects can be returned allowing standard matplotlib
          functionality to be used to augment the plot generated. See the
          :ref:`plots tutorial <Plots Tutorial>` for guidance.

    .. seealso::
        :meth:`~geopyv.mesh.MeshBase.inspect`

    """
    # Load image.
    image_path = data["images"]["f_img"]
    image = cv2.imread(image_path, cv2.IMREAD_COLOR)
    image_gs = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    image_gs = cv2.GaussianBlur(image_gs, ksize=(5, 5), sigmaX=1.1, sigmaY=1.1)

    # Load data.
    nodes = data["nodes"]
    elements = data["elements"]

    # Extract variables from data.
    x = []
    y = []
    value = []
    for n in nodes:
        x.append(n[0])
        y.append(n[1])
    x = np.asarray(x)
    y = np.asarray(y)
    value = np.asarray(value)

    # Plot setup.
    try:
        platform = sys.platform
        if platform == "linux" or platform == "linux2" or platform == "darwin":
            split = "/"
        elif platform == "win32":
            split = r"\\"
        f_img = data["images"]["f_img"][
            ([(m.end(0)) for m in re.finditer(split, data["images"]["f_img"])][-1]) :
        ]
        title = "Inspect mesh: f_img = {f_img}".format(f_img=f_img)
    except Exception:
        log.warning("Could not extract image names for plot.")
        title = "Inspect mesh:"

    # Triangulation.
    _, x_p, y_p = gp.geometry.utilities.plot_triangulation(
        elements, x, y, data["mesh_order"]
    )

    # Plot figure.
    fig, ax = plt.subplots(num=title)
    for i in range(np.shape(x_p)[0]):
        ax.plot(x_p[i], y_p[i], color="b", alpha=1.0, linewidth=1.0)
    ax.imshow(
        image,
        cmap="gist_gray",
        interpolation="nearest",
        aspect="equal",
    )
    details = r"{nodes} nodes; {elements} elements".format(
        nodes=np.shape(nodes)[0], elements=np.shape(elements)[0]
    )
    image_size = np.shape(image)
    ax.text(
        image_size[1] / 2,
        image_size[0] * 1.05,
        details,
        horizontalalignment="center",
    )
    ax.set_axis_off()
    plt.tight_layout()

    # Save
    if save is not None:
        plt.savefig(save, dpi=600)

    # Show or close.
    if show is True:
        plt.show(block=block)
    else:
        plt.close(fig)

    return fig, ax


def trace_particle(
    data,
    quantity,
    component,
    obj_type,
    imshow,
    colorbar,
    ticks,
    alpha,
    axis,
    xlim,
    ylim,
    show,
    block,
    save,
):
    title = "Path; variable = {variable}".format(variable=quantity)
    labels = [
        r"$u$ ($px$)",
        r"$v$ ($px$)",
        r"$du/dx$ ($-$)",
        r"$dv/dx$ ($-$)",
        r"$du/dy$ ($-$)",
        r"$dv/dy$ ($-$)",
        r"$d^2u/dx^2$ ($-$)",
        r"$d^2v/dx^2$ ($-$)",
        r"$d^2u/dxdy$ ($-$)",
        r"$d^2v/dxdy$ ($-$)",
        r"$d^2u/dy^2$ ($-$)",
        r"$d^2v/dy^2$ ($-$)",
    ]

    fig, ax = plt.subplots(num=title)

    if obj_type == "Particle":
        points = data["results"]["coordinates"].reshape(-1, 1, 2)
        segments = np.concatenate([points[:-1], points[1:]], axis=1)
        values = np.diff(data["results"][quantity][:, component], axis=0)
        norm = plt.Normalize(values.min(), values.max())
        lc = LineCollection(segments, cmap="viridis", norm=norm)
        lc.set_array(values)
        lines = ax.add_collection(lc)

    elif obj_type == "Field":
        values = np.empty(
            (
                len(data["particles"]),
                len(data["particles"][0]["results"][quantity]) - 1,
            )
        )
        if data["number_images"] > 1:
            segments = np.empty(
                (
                    len(data["particles"]),
                    data["number_images"] - 1,
                    2,
                    2,
                )
            )
        else:
            segments = np.empty((len(data["particles"]), 2, 2))
        for i in range(len(data["particles"])):
            values[i] = np.diff(
                data["particles"][i]["results"][quantity][:, component], axis=0
            )
            points = data["particles"][i]["results"]["coordinates"].reshape(-1, 1, 2)
            segments[i] = np.concatenate([points[:-1], points[1:]], axis=1)
        values = values.flatten()
        norm = plt.Normalize(values.min(), values.max())
        lc = LineCollection(segments.reshape(-1, 2, 2), cmap="viridis", norm=norm)
        lc.set_array(values.flatten())
        lines = ax.add_collection(lc)

    # Show image in background.
    if imshow is True:
        image = cv2.imread(data["image_0"], cv2.IMREAD_COLOR)
        image_gs = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        image_gs = cv2.GaussianBlur(image_gs, ksize=(5, 5), sigmaX=1.1, sigmaY=1.1)
        plt.imshow(image_gs, cmap="gray")
    else:
        ax.set_aspect("equal", "box")

    if colorbar is True:
        fig.colorbar(lines, label=labels[component])

        # Limit control.
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    # Save.
    if save is not None:
        plt.savefig(save, dpi=600)

    # Show or close.
    if show is True:
        plt.show(block=block)
    else:
        plt.close(fig)

    return fig, ax


def history_particle(data, quantity, components, xlim, ylim, show, block, save):
    title = "History; {quantity}".format(quantity=quantity)
    labels = [
        r"$u$ ($px$)",
        r"$v$ ($px$)",
        r"$du/dx$ ($-$)",
        r"$dv/dx$ ($-$)",
        r"$du/dy$ ($-$)",
        r"$dv/dy$ ($-$)",
        r"$d^2u/dx^2$ ($-$)",
        r"$d^2v/dx^2$ ($-$)",
        r"$d^2u/dxdy$ ($-$)",
        r"$d^2v/dxdy$ ($-$)",
        r"$d^2u/dy^2$ ($-$)",
        r"$d^2v/dy^2$ ($-$)",
    ]

    fig, ax = plt.subplots(num=title)
    if components is None:
        components = range(np.shape(data["results"][quantity])[1])
    for component in components:
        ax.plot(
            range(np.shape(data["results"][quantity])[0]),
            data["results"][quantity][:, component],
            label=labels[component],
        )

    # General formatting.
    # Legend.
    plt.legend(bbox_to_anchor=(1.05, 0.5), loc="center left", borderaxespad=0)

    # Logscale.
    ax.set_xscale("linear")
    ax.set_yscale("linear")

    # Limit control.
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    # Axis labels.
    ax.set_xlabel(r"Image Number, $i$ ($-$)")
    ax.set_ylabel(r"Value")

    # Save.
    if save is not None:
        plt.savefig(save, bbox_inches="tight", dpi=600)

    # Show or close.
    if show is True:
        plt.show(block=block)
    else:
        plt.close(fig)

    return fig, ax


def inspect_field(data, mesh, show, block, save):
    # Plot setup.
    title = "Inspect particle field"
    fig, ax = plt.subplots(num=title)

    # Load image.
    image_path = data["image_0"]
    image = cv2.imread(image_path, cv2.IMREAD_COLOR)
    image_gs = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    image_gs = cv2.GaussianBlur(image_gs, ksize=(5, 5), sigmaX=1.1, sigmaY=1.1)

    # Plot image.
    ax.imshow(
        image,
        cmap="gist_gray",
        interpolation="nearest",
        aspect="equal",
    )

    if mesh:
        # Load data.
        try:
            nodes = data["field"]["nodes"]
            elements = data["field"]["elements"]

            # Extract variables from data.
            x = []
            y = []
            value = []
            for n in nodes:
                x.append(n[0])
                y.append(n[1])
            x = np.asarray(x)
            y = np.asarray(y)
            value = np.asarray(value)

            # Triangulation
            _, x_p, y_p = gp.geometry.utilities.plot_triangulation(elements, x, y, 1)

            # Plot mesh.
            for i in range(np.shape(x_p)[0]):
                ax.plot(
                    x_p[i],
                    y_p[i],
                    color="b",
                    alpha=1.0,
                    linewidth=1.0,
                )
        except Exception:
            log.warning(
                "Mesh requested but no field mesh information "
                "exists as field was user-specified."
            )

    for i in range(len(data["field"]["coordinates"])):
        ax.scatter(
            data["field"]["coordinates"][i, 0],
            data["field"]["coordinates"][i, 1],
            c="r",
            marker="o",
        )
    details = r"{elements} particles".format(elements=np.shape(elements)[0])
    image_size = np.shape(image)
    ax.text(
        image_size[1] / 2,
        image_size[0] * 1.05,
        details,
        horizontalalignment="center",
    )
    ax.set_axis_off()
    plt.tight_layout()

    # Save
    if save is not None:
        plt.savefig(save, dpi=600)

    # Show or close.
    if show is True:
        plt.show(block=block)
    else:
        plt.close(fig)

    return fig, ax


def standard_error_validation(
    data,
    component,
    observing,
    xlim,
    ylim,
    scale,
    prev_series,
    prev_series_label,
    plot,
    show,
    block,
    save,
):
    labels = [
        r"$u$ ($px$)",
        r"$v$ ($px$)",
        r"$du/dx$ ($-$)",
        r"$dv/dx$ ($-$)",
        r"$du/dy$ ($-$)",
        r"$dv/dy$ ($-$)",
        r"$d^2u/dx^2$ ($-$)",
        r"$d^2v/dx^2$ ($-$)",
        r"$d^2u/dxdy$ ($-$)",
        r"$d^2v/dxdy$ ($-$)",
        r"$d^2u/dy^2$ ($-$)",
        r"$d^2v/dy^2$ ($-$)",
        r"$\theta$ ($^o$)",
        r"$\epsilon_{\gamma}$ ($-$)",
    ]
    colours = [
        "r",
        "b",
        "g",
        "orange",
    ]
    markers = [
        "o",
        "^",
        "s",
        "v",
    ]
    title = r"Standard error: component = {component}".format(
        component=labels[component]
    )
    fig, ax = plt.subplots(num=title)
    for i in range(len(data["applied"])):
        series = np.zeros((np.shape(data["applied"][i])[0], 2))
        for j in range(np.shape(data["applied"][i])[0]):
            if component == 12:
                series[j, 0] = (
                    180 / np.pi * abs(np.arccos(data["speckle"].data["pm"][j + 1, 2]))
                )
            elif component == 13:
                series[j, 0] = abs(data["speckle"].data["pm"][j + 1, 3])
            else:
                series[j, 0] = abs(data["speckle"].data["pm"][j + 1, component])
            if observing is not None:
                series[j, 1] = np.std(
                    np.sqrt(
                        (
                            data["applied"][i][j, :, observing]
                            - data["observed"][i][j, :, observing]
                        )
                        ** 2
                    )
                )
            else:
                series[j, 1] = np.std(
                    np.sqrt(
                        np.sum(
                            (
                                data["applied"][i][j, :, :2]
                                - data["observed"][i][j, :, :2]
                            )
                            ** 2,
                            axis=-1,
                        )
                    )
                )
        if plot == "scatter":
            ax.scatter(
                series[:, 0],
                series[:, 1],
                facecolors="none",
                edgecolors=colours[i],
                marker=markers[i],
                label=data["labels"][i],
            )
        elif plot == "line":
            ax.plot(
                series[:, 0],
                series[:, 1],
                color=colours[i],
                label=data["labels"][i],
            )
    if prev_series is not None:
        ax.plot(
            prev_series[:, 0], prev_series[:, 1], color="k", label=prev_series_label
        )
    # General formatting.
    # Legend.
    plt.legend(bbox_to_anchor=(0.05, 0.95), loc="upper left", borderaxespad=0)

    # Logscale.
    ax.set_xscale(scale)
    ax.set_yscale("log")
    ax.grid(which="both")
    ax.set_axisbelow(True)

    # Limit control.
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    # Axis labels.
    ax.set_xlabel(r"Applied warp, {}".format(labels[component]))
    if observing is not None:
        ax.set_ylabel(r"Error, $\Delta$" + labels[observing])
    else:
        ax.set_ylabel(r"Standard error, $\rho_{px}$ ($px$)")

    # Save.
    if save is not None:
        plt.savefig(save, bbox_inches="tight", dpi=600)

    # Show or close.
    if show is True:
        plt.show(block=block)
    else:
        plt.close(fig)

    return fig, ax


def mean_error_validation(
    data,
    component,
    xlim,
    ylim,
    scale,
    prev_series,
    prev_series_label,
    plot,
    show,
    block,
    save,
):
    labels = [
        r"$u$ ($px$)",
        r"$v$ ($px$)",
        r"$du/dx$ ($-$)",
        r"$dv/dx$ ($-$)",
        r"$du/dy$ ($-$)",
        r"$dv/dy$ ($-$)",
        r"$d^2u/dx^2$ ($-$)",
        r"$d^2v/dx^2$ ($-$)",
        r"$d^2u/dxdy$ ($-$)",
        r"$d^2v/dxdy$ ($-$)",
        r"$d^2u/dy^2$ ($-$)",
        r"$d^2v/dy^2$ ($-$)",
        r"$\theta$ ($^o$)",
        r"$\epsilon_{\gamma}$ ($-$)",
    ]
    colours = [
        "r",
        "b",
        "g",
        "orange",
    ]
    markers = [
        "o",
        "^",
        "s",
        "v",
    ]
    title = r"Mean error: component = {component}".format(component=labels[component])
    fig, ax = plt.subplots(num=title)
    for i in range(len(data["applied"])):
        series = np.zeros((np.shape(data["applied"][i])[0], 2))
        for j in range(np.shape(data["applied"][i])[0]):
            if component == 12:
                series[j, 0] = (
                    180 / np.pi * abs(np.arccos(data["speckle"].data["pm"][j + 1, 2]))
                )
            elif component == 13:
                series[j, 0] = abs(data["speckle"].data["pm"][j + 1, 3])
            else:
                series[j, 0] = abs(data["speckle"].data["pm"][j + 1, component])
            series[j, 1] = np.mean(
                np.sqrt(
                    np.sum(
                        (data["applied"][i][j, :, :2] - data["observed"][i][j, :, :2])
                        ** 2,
                        axis=-1,
                    )
                )
            )
        if plot == "scatter":
            ax.scatter(
                series[:, 0],
                series[:, 1],
                facecolors="none",
                edgecolors=colours[i],
                marker=markers[i],
                label=data["labels"][i],
            )
        elif plot == "line":
            ax.plot(
                series[:, 0],
                series[:, 1],
                color=colours[i],
                label=data["labels"][i],
            )
    if prev_series is not None:
        ax.plot(
            prev_series[:, 0], prev_series[:, 1], color="k", label=prev_series_label
        )
    # General formatting.
    # Legend.
    plt.legend(bbox_to_anchor=(0.05, 0.95), loc="upper left", borderaxespad=0)

    # Logscale.
    ax.set_xscale(scale)
    ax.set_yscale("log")
    ax.grid(which="both")
    ax.set_axisbelow(True)

    # Limit control.
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    # Axis labels.
    ax.set_xlabel(r"Applied warp, {}".format(labels[component]))
    ax.set_ylabel(r"Mean error, $\mu_{px}$ ($px$)")

    # Save.
    if save is not None:
        plt.savefig(save, bbox_inches="tight", dpi=600)

    # Show or close.
    if show is True:
        plt.show(block=block)
    else:
        plt.close(fig)

    return fig, ax


def spatial_error_validation(
    data,
    field_index,
    component,
    imshow,
    colorbar,
    xlim,
    ylim,
    show,
    block,
    save,
):
    labels = [
        r"$u$ ($px$)",
        r"$v$ ($px$)",
        r"$du/dx$ ($-$)",
        r"$dv/dx$ ($-$)",
        r"$du/dy$ ($-$)",
        r"$dv/dy$ ($-$)",
        r"$d^2u/dx^2$ ($-$)",
        r"$d^2v/dx^2$ ($-$)",
        r"$d^2u/dxdy$ ($-$)",
        r"$d^2v/dxdy$ ($-$)",
        r"$d^2u/dy^2$ ($-$)",
        r"$d^2v/dy^2$ ($-$)",
        r"$\theta$ ($^o$)",
        r"$\epsilon_{\gamma}$ ($-$)",
    ]

    title = r"Error: component = {component}".format(component=labels[component])
    fig, ax = plt.subplots(num=title)
    values = np.empty(
        (
            len(data["fields"][field_index].data["particles"]),
            len(
                data["fields"][field_index].data["particles"][0]["results"][
                    "coordinates"
                ]
            )
            - 1,
        )
    )
    if data["fields"][field_index].data["number_images"] > 1:
        segments = np.empty(
            (
                len(data["fields"][field_index].data["particles"]),
                data["fields"][field_index].data["number_images"] - 1,
                2,
                2,
            )
        )
    else:
        segments = np.empty((len(data["fields"][field_index].data["particles"]), 2, 2))
    for i in range(len(data["fields"][field_index].data["particles"])):
        values[i] = np.sqrt(
            np.sum(
                (
                    data["applied"][field_index][:, i, :2]
                    - data["observed"][field_index][:, i, :2]
                )
                ** 2,
                axis=-1,
            )
        )
        points = (
            data["fields"][field_index]
            .data["particles"][i]["results"]["coordinates"]
            .reshape(-1, 1, 2)
        )
        segments[i] = np.concatenate([points[:-1], points[1:]], axis=1)
    values = values.flatten()
    norm = plt.Normalize(values.min(), values.max())
    lc = LineCollection(segments.reshape(-1, 2, 2), cmap="viridis", norm=norm)
    lc.set_array(values.flatten())
    lines = ax.add_collection(lc)

    # Show image in background.
    if imshow is True:
        image = cv2.imread(
            data["fields"][field_index].data["image_0"], cv2.IMREAD_COLOR
        )
        image_gs = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        image_gs = cv2.GaussianBlur(image_gs, ksize=(5, 5), sigmaX=1.1, sigmaY=1.1)
        plt.imshow(image_gs, cmap="gray")
    else:
        ax.set_aspect("equal", "box")

    if colorbar is True:
        fig.colorbar(lines, label=labels[component])

        # Limit control.
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)

    # Save.
    if save is not None:
        plt.savefig(save, dpi=600)

    # Show or close.
    if show is True:
        plt.show(block=block)
    else:
        plt.close(fig)

    return fig, ax


def inspect_calibration(data, image_index):
    fig, ax = plt.subplots()
    frame = cv2.imread(data["file_settings"]["calibration_images"][image_index])
    print(data["calibration"]["corners"][image_index])
    print(data["calibration"]["ids"][image_index])

    # cv2.aruco.drawDetectedMarkers(frame, data["calibration"]["corners"][image_index],
    # data["calibration"]["ids"][image_index])
    ax.imshow(frame)
    ax.scatter(
        data["calibration"]["corners"][image_index][:, :, 0],
        data["calibration"]["corners"][image_index][:, :, 1],
        color="r",
    )
    plt.show()

    return fig, ax



from PySide6.QtWidgets import QApplication, QFileDialog, QWidget


class FolderSelector:
    def __init__(self):
        """

        Graphical user interface to allow the user to select a folder
        using the native file browser on the host OS.

        """
        app = QApplication.instance()
        if app is None:
            app = QApplication()
        self.widget = QWidget()

    def get_path(self, message, directory):
        """

        Method to get the path of the selected folder.

        Parameters
        ----------
        message : str
            Message to display on the file browser window header.
        directory : str
            Directory to start from.

        Returns
        -------
        path : str
            Path to selected folder.

        """
        path = QFileDialog.getExistingDirectory(self.widget, directory, message)
        return path



"""

Field module for geopyv.

"""
import logging
import numpy as np
import geopyv as gp
from geopyv.object import Object
import gmsh
from scipy.optimize import minimize_scalar
from alive_progress import alive_bar

log = logging.getLogger(__name__)


class FieldBase(Object):
    """
    Field base class to be used as a mixin.
    """

    def __init__(self):
        super().__init__(object_type="Field")
        """

        Field base class initialiser.

        """

    def inspect(self, mesh=True, show=True, block=True, save=None):
        """
        Method to show the particles and associated representative areas.

        Parameters
        ----------
        mesh : bool, optional
            Control whether the mesh is plotted.
            Defaults to True.
        show : bool, optional
            Control whether the plot is displayed.
        block : bool, optional
            Control whether the plot blocks execution until closed.
        save : str, optional
            Name to use to save plot. Uses default extension of `.png`.
        """

        # Check inputs.
        self._report(gp.check._check_type(mesh, "mesh", [bool]), "TypeError")
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        log.info("Inspecting field...")
        fig, ax = gp.plots.inspect_field(
            self.data, mesh=mesh, show=show, block=block, save=save
        )
        return fig, ax

    def trace(
        self,
        *,
        quantity="warps",
        particle_index=None,
        component=0,
        imshow=True,
        colorbar=True,
        ticks=None,
        alpha=0.75,
        axis=True,
        xlim=None,
        ylim=None,
        show=True,
        block=True,
        save=None,
    ):
        """
        Method to plot an incremental quantity along the particle position path.

        Parameters
        ----------
        quantity : str, optional
            Specifier for which metric to plot along the particle path.
        component : int, optional
            Specifier for which component of the metric to plot along the particle path.
        imshow : bool, optional
            Control whether the reference image is plotted.
            Defaults to True.
        colorbar : bool, optional
            Control whether the colour bar is plotted.
            Defaults to True.
        ticks : list, optional
            Overwrite default colourbar ticks.
            Defaults to None.
        alpha : float, optional
            Control contour opacity. Must be between 0.0-1.0.
            Defaults to 0.75.
        axis : bool, optional
            Control whether the axes are plotted.
            Defaults to True.
        xlim : array-like, optional
            Set the plot x-limits (lower_limit,upper_limit).
            Defaults to None.
        ylim : array-like, optional
            Set the plot y-limits (lower_limit,upper_limit).
            Defaults to None.
        show : bool, optional
            Control whether the plot is displayed.
            Defaults to True.
        block : bool, optional
            Control whether the plot blocks execution until closed.
            Defaults to False.
        save : str, optional
            Name to use to save plot. Uses default extension of `.png`.
        """

        # Check if solved.
        if self.data["solved"] is not True:
            log.error(
                "Particle not yet solved therefore no convergence data to plot. "
                "First, run :meth:`~geopyv.particle.Particle.solve()` to solve."
            )
            raise ValueError(
                "Particle not yet solved therefore no convergence data to plot. "
                "First, run :meth:`~geopyv.particle.Particle.solve()` to solve."
            )
        # Check input.
        self._report(gp.check._check_type(quantity, "quantity", [str]), "TypeError")
        if quantity:
            self._report(
                gp.check._check_value(
                    quantity,
                    "quantity",
                    [
                        "coordinates",
                        "warps",
                        "volumes",
                        "stresses",
                    ],
                ),
                "ValueError",
            )
        self._report(gp.check._check_type(component, "component", [int]), "TypeError")
        self._report(
            gp.check._check_index(
                component,
                "component",
                1,
                self.data["particles"][0]["results"][quantity],
            ),
            "IndexError",
        )
        self._report(gp.check._check_type(imshow, "imshow", [bool]), "TypeError")
        self._report(gp.check._check_type(colorbar, "colorbar", [bool]), "TypeError")
        types = [tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(ticks, "ticks", types), "TypeError")
        check = gp.check._check_type(alpha, "alpha", [float])
        if check:
            try:
                alpha = float(alpha)
                self._report(gp.check._conversion(alpha, "alpha", float), "Warning")
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_range(alpha, "alpha", 0.0, 1.0), "ValueError")
        self._report(gp.check._check_type(axis, "axis", [bool]), "TypeError")
        types = [tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(xlim, "xlim", types), "TypeError")
        if xlim is not None:
            self._report(gp.check._check_dim(xlim, "xlim", 1), "ValueError")
            self._report(gp.check._check_axis(xlim, "xlim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(ylim, "ylim", types), "TypeError")
        if ylim is not None:
            self._report(gp.check._check_dim(ylim, "ylim", 1), "ValueError")
            self._report(gp.check._check_axis(ylim, "ylim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        if particle_index is None:
            log.info("Tracing field...")
            obj_type = "Field"
            data = self.data
        else:
            log.info("Tracing particle...")
            obj_type = "Particle"
            data = self.data["particles"][particle_index]
        fig, ax = gp.plots.trace_particle(
            data=data,
            quantity=quantity,
            component=component,
            obj_type=obj_type,
            imshow=imshow,
            colorbar=True,
            ticks=ticks,
            alpha=alpha,
            axis=axis,
            xlim=xlim,
            ylim=ylim,
            show=show,
            block=block,
            save=save,
        )
        return fig, ax

    def history(
        self,
        particle_index,
        quantity="warps",
        components=None,
        xlim=None,
        ylim=None,
        show=True,
        block=True,
        save=None,
    ):
        """
        Method to plot particle time history.

        Parameters
        ----------
        quantity : str, optional
            Specifier for which metric to plot along the particle path.
        component : int, optional
            Specifier for which components of the metric to plot.
        xlim : array-like, optional
            Set the plot x-limits (lower_limit,upper_limit).
            Defaults to None.
        ylim : array-like, optional
            Set the plot y-limits (lower_limit,upper_limit).
            Defaults to None.
        show : bool, optional
            Control whether the plot is displayed.
            Defaults to True.
        block : bool, optional
            Control whether the plot blocks execution until closed.
            Defaults to False.
        save : str, optional
            Name to use to save plot. Uses default extension of `.png`.
        """

        # Check if solved.
        if self.data["solved"] is not True:
            log.error(
                "Particle not yet solved therefore no convergence data to plot. "
                "First, run :meth:`~geopyv.particle.Particle.solve()` to solve."
            )
            raise ValueError(
                "Particle not yet solved therefore no convergence data to plot. "
                "First, run :meth:`~geopyv.particle.Particle.solve()` to solve."
            )

        # Check input.
        self._report(
            gp.check._check_type(quantity, "quantity", [str, type(None)]), "TypeError"
        )
        if quantity:
            self._report(
                gp.check._check_value(
                    quantity,
                    "quantity",
                    [
                        "coordinates",
                        "warps",
                        "volumes",
                        "stresses",
                    ],
                ),
                "ValueError",
            )
        self._report(
            gp.check._check_type(components, "components", [list, type(None)]),
            "TypeError",
        )
        # if components:
        #    for component in components:
        #        self._report(
        #            gp.check._check_index(
        #                component, "component", 1, self.data["results"][quantity]
        #            ),
        #            "IndexError",
        #        )
        types = [tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(xlim, "xlim", types), "TypeError")
        if xlim is not None:
            self._report(gp.check._check_dim(xlim, "xlim", 1), "ValueError")
            self._report(gp.check._check_axis(xlim, "xlim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(ylim, "ylim", types), "TypeError")
        if ylim is not None:
            self._report(gp.check._check_dim(ylim, "ylim", 1), "ValueError")
            self._report(gp.check._check_axis(ylim, "ylim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        fig, ax = gp.plots.history_particle(
            data=self.data["particles"][particle_index],
            quantity=quantity,
            components=components,
            xlim=xlim,
            ylim=ylim,
            show=show,
            block=block,
            save=save,
        )
        return fig, ax

    def _report(self, msg, error_type):
        if msg and error_type != "Warning":
            log.error(msg)
        elif msg and error_type == "Warning":
            log.warning(msg)
            return True
        if error_type == "ValueError" and msg:
            raise ValueError(msg)
        elif error_type == "TypeError" and msg:
            raise TypeError(msg)
        elif error_type == "IndexError" and msg:
            raise IndexError(msg)


class Field(FieldBase):
    def __init__(
        self,
        *,
        series=None,
        target_particles=1000,
        track=True,
        boundary_obj=None,
        exclusion_objs=[],
        coordinates=None,
        volumes=None,
        stresses=np.zeros(6),
    ):
        """
        Initialisation of geopyv field object.

        Parameters
        ----------
        series : gp.sequence.Sequence object or gp.mesh.Mesh object
            The base series for field object interpolation.
        target_particles : int, optional
            Target number of particles. Defaults to a value of 1000.
        track : bool, optional
            Boolean to specify if particles should move or remain static.
            True: move (Lagrangian), False: static (Eularian). Defaults to True.
        boundary : numpy.ndarray (N,2), optional
            Array of coordinates to define the particle auto-distribution mesh
            boundary.
        exclusions : list, optional
            List to define the particle auto-distribution mesh exclusions.
            Shape of `numpy.ndarray` (N,2).
        coordinates : numpy.ndarray (N,2), optional
            Array of coordinates to define the initial particle positions.
        volumes : numpy.ndarray (N,), optional
            Array of volumes for particle representation.
            Defaults to np.ones(N) i.e. measure of volumetric strain.
        stresses : numpy.ndarray (N,6) or (2,6) or (6,), optional.
            Stress state specification. If:
            (N,6) - individual particle specification (where N is the number
                    of particles).
            (2,6) - linearly varying stress field with depth (1st row and 2nd
                    row vertically highest and lowest respectively, according
                    to the boundary definition.
            (6,)  - uniform stress field.
            Defaults to np.zeros(6).

        Note ::
        Two kwargs groups for particle distribution:
        if coordinates is not None:
            1. User-defined : coordinates, volumes.
        else:
            2. Auto-distributed: boundary, exclusions, target_particles.

        Attributes
        ----------
        data : dict
            Data object containing all settings and results.
            See the data structure :ref:`here <field_data_structure>`.
        solved : bool
            Boolean to indicate if the field has been solved.

        """
        self._initialised = False
        _auto_distribute = True

        # Check inputs.
        types = [
            gp.sequence.Sequence,
            gp.sequence.SequenceResults,
            gp.mesh.Mesh,
            gp.mesh.MeshResults,
        ]
        self._report(gp.check._check_type(series, "series", types), "TypeError")
        check = gp.check._check_type(target_particles, "target_particles", [int])
        if check:
            try:
                target_particles = int(target_particles)
                self._report(
                    gp.check._conversion(target_particles, "target_particles", int),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(target_particles, "target_particles", 1), "ValueError"
        )
        self._report(
            gp.check._check_type(
                boundary_obj,
                "boundary_obj",
                [gp.geometry.region.Circle, gp.geometry.region.Path, type(None)],
            ),
            "TypeError",
        )
        check = gp.check._check_type(exclusion_objs, "exclusion_objs", [list])
        if check:
            try:
                exclusion_objs = list(exclusion_objs)
                self._report(
                    gp.check._conversion(exclusion_objs, "exclusion_objs", list, False),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        for exclusion_obj in exclusion_objs:
            self._report(
                gp.check._check_type(
                    exclusion_obj,
                    "exclusion_obj",
                    [gp.geometry.region.Circle, gp.geometry.region.Path, type(None)],
                ),
                "TypeError",
            )
        check = gp.check._check_type(
            coordinates, "coordinates", [np.ndarray, type(None)]
        )
        if check:
            try:
                coordinates = np.asarray(coordinates)
                self._report(
                    gp.check._conversion(coordinates, "coordinates", np.ndarray, False),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        if coordinates is not None:
            self._report(
                gp.check._check_dim(coordinates, "coordinates", 2), "ValueError"
            )
            self._report(
                gp.check._check_axis(coordinates, "coordinates", 1, [2]), "ValueError"
            )
        check = gp.check._check_type(volumes, "volumes", [np.ndarray, type(None)])
        if check:
            try:
                volumes = np.asarray(volumes)
                self._report(
                    gp.check._conversion(volumes, "volumes", np.ndarray, False),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        if volumes:
            self._report(gp.check._check_dim(volumes, "volumes", 1), "ValueError")
            self._report(
                gp.check._check_axis(volumes, "volumes", 0, [np.shape(coordinates)[0]]),
                "ValueError",
            )
        check = gp.check._check_type(stresses, "stresses", [np.ndarray])
        if check:
            try:
                stresses = np.asarray(stresses)
                self._report(
                    gp.check._conversion(stresses, "stresses", np.ndarray, False),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        if coordinates is not None:
            self._report(
                gp.check._check_axis(
                    stresses, "stresses", 0, [6, 2, np.shape(coordinates)[0]]
                ),
                "ValueError",
            )
        else:
            self._report(
                gp.check._check_axis(stresses, "stresses", 0, [6, 2]), "ValueError"
            )

        self._report(gp.check._check_type(track, "track", [bool]), "TypeError")

        # Store.
        if series.data["type"] == "Sequence":
            self._series_type = "Sequence"
            if "file_settings" in series.data:
                if series.data["file_settings"]["save_by_reference"]:
                    mesh_no = np.shape(series.data["meshes"])[0]
                    _meshes = np.empty(mesh_no, dtype=object)
                    with alive_bar(
                        mesh_no, dual_line=True, bar="blocks", title="Loading meshes..."
                    ) as bar:
                        for i in range(mesh_no):
                            _meshes[i] = gp.io.load(
                                filename=series.data["file_settings"]["mesh_dir"]
                                + series.data["meshes"][i],
                                verbose=False,
                            ).data
                            bar()
                else:
                    _meshes = series.data["meshes"]
            else:
                _meshes = series.data["meshes"]
            series.data["meshes"] = _meshes
            series.data["file_settings"]["save_by_reference"] = False
            mesh_0 = series.data["meshes"][0]
            self._number_images = np.shape(series.data["meshes"])[0] + 1
        else:
            self._series_type = "Mesh"
            mesh_0 = series.data
            self._number_images = 2
        self._image_0 = mesh_0["images"]["f_img"]
        self._series = series
        self._track = track
        self.solved = False
        self._unsolvable = False

        if coordinates is not None:
            _auto_distribute = False
            image = gp.image.Image(self._image_0)
            for coord in coordinates:
                if (
                    coord[0] > np.shape(image.image_gs)[1]
                    or coord[0] < 0
                    or coord[1] > np.shape(image.image_gs)[0]
                    or coord[1] < 0
                ):
                    log.error(
                        (
                            "User-specified coordinate {value} "
                            "outside image boundary."
                        ).format(value=coord)
                    )
                    raise ValueError(
                        (
                            "User-specified coordinate {value} "
                            "outside image boundary."
                        ).format(value=coord)
                    )
            del image
            if volumes is None:
                volumes = np.ones(np.shape(coordinates)[0])
            self._target_particles = np.shape(coordinates)[0]
        else:
            self._target_particles = target_particles

        self.data = {
            "type": "Field",
            "solved": self.solved,
            "series_type": self._series_type,
            "number_images": self._number_images,
            "track": self._track,
            "target_particles": self._target_particles,
            "image_0": self._image_0,
        }

        # Particle distribution.
        if _auto_distribute is True:
            self._boundary_obj = mesh_0["boundary_obj"]
            self._exclusion_objs = mesh_0["exclusion_objs"]
            self._size_lower_bound = mesh_0["size_lower_bound"]
            self._size_upper_bound = mesh_0["size_upper_bound"]
            (
                self._borders,
                self._segments,
                self._curves,
                self._mask,
            ) = gp.geometry.meshing._define_RoI(
                gp.image.Image(self._image_0), self._boundary_obj, self._exclusion_objs
            )
            # Initialize gmsh if not already initialized.
            if gmsh.isInitialized() == 0:
                gmsh.initialize()
            gmsh.option.setNumber("General.Verbosity", 2)
            log.info(
                "Generating mesh using gmsh with approximately {n} particles.".format(
                    n=self._target_particles
                )
            )
            self._initial_mesh()
            self._update_mesh()
            gmsh.finalize()
            self._distribute_particles()
            self._stress_state(stresses)
            log.info(
                "Field generated with {p} particles.".format(p=len(self._coordinates))
            )
            self._field = {
                "nodes": self._nodes,
                "elements": self._elements,
                "coordinates": self._coordinates,
                "volumes": self._volumes,
                "stresses": self._stresses,
            }
        else:
            self._coordinates = coordinates
            self._volumes = volumes
            self._stress_state(stresses)
            log.info("Using user-specified field.")
            self._field = {
                "coordinates": self._coordinates,
                "volumes": self._volumes,
                "stresses": self._stresses,
            }

        self.data.update({"field": self._field})
        self._initialised = True

    def _initial_mesh(self):
        """

        Private method to optimize the element size to generate
        approximately the desired number of elements.

        """

        def f(size):
            return self._uniform_remesh(
                size,
                self._borders,
                self._segments,
                self._curves,
                self._target_particles,
                self._size_lower_bound,
                self._size_upper_bound,
            )

        minimize_scalar(
            f,
            bounds=(self._size_lower_bound, self._size_upper_bound),
            method="bounded",
        )

    def _update_mesh(self):
        """

        Private method to update the mesh variables.

        """
        (
            _,
            nc,
            _,
        ) = gmsh.model.mesh.getNodes()  # Extracts: node coordinates.
        _, _, ent = gmsh.model.mesh.getElements(dim=2)  # Extracts: element node tags.
        self._nodes = np.column_stack(
            (nc[0::3], nc[1::3])
        )  # Nodal coordinate array (x,y).
        self._elements = np.reshape(
            (np.asarray(ent) - 1).flatten(), (-1, 3)
        )  # Element connectivity array.

    def _distribute_particles(self):
        self._coordinates = np.mean(self._nodes[self._elements[:, :3]], axis=1)
        M = np.ones((len(self._elements[:, :3]), 3, 3))
        M[:, 1] = self._nodes[self._elements[:, :3]][:, :, 0]
        M[:, 2] = self._nodes[self._elements[:, :3]][:, :, 1]
        self._volumes = abs(0.5 * np.linalg.det(M))

    def solve(
        self,
        *,
        model=None,
        state=None,
        parameters=None,
    ):
        """
        Method to solve for the field.

        Returns
        -------
        solved : bool
            Boolean to indicate if the particle instances have been solved.
        """
        particle_no = np.shape(self._coordinates)[0]
        self._particles = np.empty(particle_no, dtype=dict)
        self._vol_totals = np.zeros(self.data["number_images"])
        with alive_bar(
            particle_no, dual_line=True, bar="blocks", title="Solving particles..."
        ) as bar:
            for i in range(particle_no):
                particle = gp.particle.Particle(
                    series=self._series,
                    coordinate_0=self._coordinates[i],
                    volume_0=self._volumes[i],
                    stress_0=self._stresses[i],
                    track=self._track,
                )
                _particle_solved = particle.solve(
                    model=model, state=state, parameters=parameters
                )
                if _particle_solved is False:
                    self._unsolvable = True
                    self.data["unsolvable"] = self._unsolvable
                    return self.solved
                self._particles[i] = particle.data
                self._vol_totals += self._particles[i]["results"]["volumes"]
                del particle
                bar()

        if model is not None:
            self._works = np.zeros(self.data["number_images"])
            for i in range(particle_no):
                self._works += self._particles[i]["results"]["works"]
            self.data.update({"works": self._works})

        self.data.update({"particles": self._particles, "volumes": self._vol_totals})
        self.solved = True
        self.data["solved"] = self.solved
        return self.solved

    def _stress_state(self, stresses):
        """
        Private method to define the initial stress state.
        """
        if np.shape(stresses)[0] == np.shape(self._coordinates)[0]:
            self._stresses = stresses
        elif np.shape(stresses)[0] == 2:
            self._stresses = stresses[0] + (stresses[1] - stresses[0]) * (
                (
                    (self._coordinates[:, 1] - np.min(self._boundary._boundary[:, 1]))
                    / (
                        np.max(self._boundary._boundary[:, 1])
                        - np.min(self._boundary._boundary[:, 1])
                    )
                )[:, np.newaxis]
            )
        else:
            self._stresses = np.tile(stresses, (np.shape(self._coordinates)[0], 1))

    @staticmethod
    def _uniform_remesh(
        size,
        boundary,
        segments,
        curves,
        target_particles,
        size_lower_bound,
        size_upper_bound,
    ):
        """

        Private method to create the initial mesh.

        Parameters
        ----------
        size : int
            Target size of elements.
        boundary : `numpy.ndarray` (Nx,Ny)
            Array of coordinates to define the mesh boundary.
        segments : `numpy.ndarray` (Nx,Ny)
            Array of segments for gmsh mesh generation.
        curves : `numpy.ndarray` (Nx,Ny)
            Array of curves for gmsh mesh generation.
        target_particles : int
            Target number of nodes.
        size_lower_bound : int
            Lower bound on element size.


        Returns
        -------
        error : int
            Error between target and actual number of nodes.

        """
        # Make mesh.
        gmsh.model.add("base")  # Create model.

        # Add points.
        for i in range(np.shape(boundary)[0]):
            gmsh.model.occ.addPoint(boundary[i, 0], boundary[i, 1], 0, size, i)

        # Add line segments.
        for i in range(np.shape(segments)[0]):
            gmsh.model.occ.addLine(segments[i, 0], segments[i, 1], i)

        # Add curves.
        for i in range(len(curves)):
            gmsh.model.occ.addCurveLoop(curves[i], i)
        curve_indices = list(np.arange(len(curves), dtype=np.int32))

        # Create surface.
        gmsh.model.occ.addPlaneSurface(curve_indices, 0)

        # Generate mesh.
        gmsh.option.setNumber("Mesh.MeshSizeMin", size_lower_bound)
        gmsh.option.setNumber("Mesh.MeshSizeMax", size_upper_bound)
        gmsh.model.occ.synchronize()
        gmsh.model.mesh.generate(2)
        gmsh.model.mesh.optimize()
        gmsh.model.mesh.setOrder(1)

        # Get mesh topology.
        _, _, ent = gmsh.model.mesh.getElements(dim=2)
        elements = np.reshape((np.asarray(ent) - 1).flatten(), (-1, 3))
        error = (np.shape(elements)[0] - target_particles) ** 2
        return error


class FieldResults(FieldBase):
    """

    FieldResults class for geopyv.

    Parameters
    ----------
    data : dict
        geopyv data dict from Field object.


    Attributes
    ----------
    data : dict
        geopyv data dict from Field object.

    """

    def __init__(self, data):
        """Initialisation of geopyv FieldResults class."""
        self.data = data



"""

Utilities module for geopyv.

"""
import numpy as np


def area_to_length(area):
    """

    Function that returns a characteristic length given
    an element area, based on an equilateral triangle.

    Parameters
    ----------
    area : float
        Element area.


    Returns
    -------
    length : float
        Characteristic length.

    """
    length = np.sqrt(4 * abs(area) / np.sqrt(3))
    return length


def plot_triangulation(elements, x, y, mesh_order):
    """

    Method to compute a first order triangulation from a
    second order element connectivity array and coordinates.

    Parameters
    ----------
    elements : np.ndarray (Nx, 6)
        Element connectivity array.
    x : np.ndarray (Nx, 1)
        Horizontal coordinate array.
    y : np.ndarray (Nx, 1)
        Vertical coordinate array.


    Returns
    -------
    mesh_triangulation : np.ndarray (Nx, 7)
        Mesh triangulation array for plot purposes forming
        closed triangles.
    x_p : np.ndarray (Nx, 1)
        Horizontal coordinate of triangle vertices.
    y_p : np.ndarray (Nx, 1)
        Vertical coordinate of triangle vertices.

    """
    if mesh_order == 1:
        x_p = x[elements[:, [0, 1, 2, 0]]]
        y_p = y[elements[:, [0, 1, 2, 0]]]
        mesh_triangulation = elements
    elif mesh_order == 2:
        x_p = x[elements[:, [0, 3, 1, 4, 2, 5, 0]]]
        y_p = y[elements[:, [0, 3, 1, 4, 2, 5, 0]]]
        mesh_triangulation = elements[
            :, [[0, 3, 5], [1, 3, 4], [2, 4, 5], [3, 4, 5]]
        ].reshape(-1, 3)

    return mesh_triangulation, x_p, y_p


def PolyArea(pts):
    """

    A function that returns the area of the input polygon.

    Parameters
    ----------
    pts : `numpy.ndarray` (Nx,2)
        Clockwise/anti-clockwise ordered coordinates.

    """
    x = pts[:, 0]
    y = pts[:, 1]
    return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))



"""

Image module for geopyv.

"""
import logging
import cv2
import numpy as np
import geopyv as gp
import os
from geopyv.object import Object

log = logging.getLogger(__name__)


class Image(Object):
    def __init__(self, filepath=None, border=20):
        super().__init__(object_type="Image")
        """

        Image class for geopyv Image object.

        Parameters
        ----------
        filepath : str
            Path to image.
        border : int
            Bi-quintic B-spline interpolation border. Defaults to a value of 20.


        Attributes
        ----------
        image_gs : `numpy.ndarray` (Nx, Ny)
            2D array containing grayscale image data of type `int`, generated by OpenCV
            function :py:func:`~cv2.cvtColor`.
        QCQT : `numpy.ndarray` (6*Nx, 6*Ny)
            2D array containing precomputed bi-quintic B-spline interpolation matrices
            of type `float`, computed by :meth:`~_get_QCQT`.

        """
        log.debug("Initialising geopyv Image object at {}.".format(filepath))
        if filepath is None:
            directory = os.getcwd()
            dialog = gp.gui.selectors.image.ImageSelector()
            filepath = dialog.get_path(directory, "Select image")
        elif os.path.exists(filepath):
            self.filepath = filepath
            self.border = np.int_(border)
            self._get_image_gs()
            self._get_QCQT()
            log.debug("Initialised geopyv Image object at {}.".format(filepath))
        else:
            log.error("Image does not exist at the path supplied:\n{}".format(filepath))
            raise FileExistsError(
                "Image does not exist at the path supplied:\n{}".format(filepath)
            )

    def _get_image_gs(self):
        r"""

        Private method to convert colour image to grayscale image for image processing
        purposes using OpenCV functions `imread` and `cvtColor`, and apply Gaussian
        prefilter with 5x5 kernel after Pan et al. (2013), assuming
        :math:`\sigma_{x} = \sigma_{y} = 0.3 \cdot (0.5 \cdot (k-1)-1) + 0.8 = 1.1`.

        """
        image = cv2.imread(self.filepath, cv2.IMREAD_COLOR)
        self.image_gs = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        self.image_gs = cv2.GaussianBlur(
            self.image_gs, ksize=(5, 5), sigmaX=1.1, sigmaY=1.1
        )

    def _get_C(self):
        r"""

        Private method to calculate the bi-quintic B-spline coefficient matrix
        :math:`\mathbf{C}` for image intensity interpolation.

        """
        # Pad the grayscale image.
        image_gs_pad = cv2.copyMakeBorder(
            self.image_gs,
            self.border,
            self.border,
            self.border,
            self.border,
            cv2.BORDER_REPLICATE,
        )

        # Define the quintic B-spline kernel.
        k = np.array(
            [1 / 120, 13 / 60, 11 / 20, 13 / 60, 1 / 120, 0],
            order="F",
        )

        # Create initial C array.
        C = np.zeros(
            (np.shape(image_gs_pad)[0], np.shape(image_gs_pad)[1]),
            dtype=complex,
            order="F",
        )

        # Work through rows first.
        # Setup kernel vector.
        kernel_x = np.zeros((np.shape(image_gs_pad)[1]), dtype=complex)
        kernel_x[0:3] = k[3:]
        kernel_x[-3:] = k[0:3]
        kernel_x = np.fft.fft(kernel_x)

        # FFT across rows.
        for i in np.arange(0, np.shape(image_gs_pad)[0], 1):
            C[i, :] = np.fft.ifft(np.fft.fft(image_gs_pad[i, :]) / kernel_x)

        # Now work through columns.
        # Setup kernel vector.
        kernel_y = np.zeros((np.shape(image_gs_pad)[0]), dtype=complex)
        kernel_y[0:3] = k[3:]
        kernel_y[-3:] = k[0:3]
        kernel_y = np.fft.fft(kernel_y)

        # FFT along columns.
        for j in np.arange(0, np.shape(image_gs_pad)[1], 1):
            C[:, j] = np.fft.ifft(np.fft.fft(C[:, j]) / kernel_y)

        # Discard the imaginary part.
        C = np.real(C)

        return C

    def _get_QCQT(self):
        r"""

        Private method to precompute the :math:`\mathbf{Q} \cdot
        \mathbf{C}_\left(i-2:i+3, j-2:j+3\right) \cdot \mathbf{Q^T}`
        matrix for the image for use in the bi-quintic B-spline image
        intensity interpolation process.

        """
        # Create QK array.
        Q = np.asarray(
            (
                [1 / 120, 13 / 60, 11 / 20, 13 / 60, 1 / 120, 0],
                [-1 / 24, -5 / 12, 0, 5 / 12, 1 / 24, 0],
                [1 / 12, 1 / 6, -1 / 2, 1 / 6, 1 / 12, 0],
                [-1 / 12, 1 / 6, 0, -1 / 6, 1 / 12, 0],
                [1 / 24, -1 / 6, 1 / 4, -1 / 6, 1 / 24, 0],
                [-1 / 120, 1 / 24, -1 / 12, 1 / 12, -1 / 24, 1 / 120],
            ),
            np.float64,
            order="F",
        )

        # Create transpose of Q array.
        QT = np.ascontiguousarray(np.transpose(Q), np.float64)

        # Call C++ extension to perform computations.
        C = self._get_C()
        self.QCQT = gp._image_extensions._QCQT(
            Q, QT, C, self.image_gs.shape, self.border
        )



import logging
import numpy as np
import os

log = logging.getLogger(__name__)


def _check_type(parameter, name, types):
    """

    Standard function for checking types of input parameter.

    Parameters
    ----------
    parameter :
        The input to check.
    types:
        The types expected.

    Returns
    -------
    msg :

    """

    if type(parameter) not in types:
        if len(types) > 1:
            type_str = ", ".join([val.__name__ for val in types])
            splitpoint = type_str.rfind(",")
            type_str = type_str[:splitpoint] + " or" + type_str[splitpoint + 1 :]
        else:
            type_str = types[0]
        return (
            "`{name}` kwarg type invalid. "
            "Expected a {types}, but got a {parameter_type}."
        ).format(
            name=name,
            types=type_str,
            parameter_type=type(parameter).__name__,
        )


def _check_index(index, name, axis, array):
    """

    Private function for checking input index against array bounds.

    Parameters
    ----------
    index: int
        The input index.
    name:
        The parameter name.
    axis:
        The axis to check for.
    array:
        The array to check for.

    Returns:
    None if check passed.
    Error message if check failed.
    """

    if index < 0 or index >= np.shape(array)[axis]:
        return (
            "`{name}` {value} is out of bounds " "for axis {axis} with size {maximum}"
        ).format(
            name=name,
            value=index,
            axis=axis,
            maximum=np.shape(array)[axis],
        )


def _check_value(parameter, name, values):
    """

    Private function for checking parameter value.

    Parameters
    ----------
    parameter :
    name :
    values :
    """

    if parameter not in values:
        if len(values) > 1:
            value_str = ", ".join(values)
            splitpoint = value_str.rfind(",")
            value_str = value_str[:splitpoint] + " or" + value_str[splitpoint + 1 :]
        else:
            value_str = values[0]
        return (
            "`{name}` kwarg value invalid. "
            "Expected a value from {values}, but got {value}."
        ).format(
            name=name,
            values=value_str,
            value=parameter,
        )


def _check_range(parameter, name, lb, ub=None):
    """

    Private function for checking parameter value against a specified range.

    Parameters
    ----------

    """

    if ub:
        if parameter < lb or parameter > ub:
            return ("`{name}` kwarg value {value} out of range {lb}-{ub}.").format(
                name=name, value=parameter, lb=lb, ub=ub
            )
    else:
        if parameter < lb:
            return (
                "`{name}` kwarg value {value} out of range {ub}>{name}>{lb}."
            ).format(name=name, value=parameter, lb=lb, ub=ub)


def _check_axis(array, name, axis, values):
    """

    Private function for checking axis size.

    """

    if np.shape(array)[axis] not in values:
        if len(values) > 1:
            values_str = ", ".join(values)
            splitpoint = values_str.rfind(",")
            values_str = values_str[:splitpoint] + " or" + values_str[splitpoint + 1 :]
        else:
            values_str = values[0]
        return (
            "`{name}` kwarg axis {axis} size invalid. "
            "Expected {values}, but got {size}."
        ).format(name=name, axis=axis, values=values_str, size=np.shape(array)[axis])


def _check_dim(array, name, value):
    """

    Private function for checking array dimensions.

    """

    if np.asarray(array).ndim != value:
        if value != 1 or (value == 1 and np.shape(np.asarray(array))[1] != 1):
            return (
                "`{name}` kwarg dimensions are invalid. "
                "Expected {value}D-array, but got a {dim}D-array"
            ).format(
                name=name,
                value=value,
                dim=array.ndim,
            )


def _check_comp(parameter1, name1, parameter2, name2):
    if parameter1 > parameter2:
        return "`{name1}`>`{name2}`: {value1}>{value2}.".format(
            name1=name1,
            name2=name2,
            value1=parameter1,
            value2=parameter2,
        )


def _check_path(path, name):
    if os.path.isdir(path) is False:
        return "`{name}` does not exist at the path supplied:\n{path}".format(
            name=name,
            path=path,
        )


def _check_character(parameter, character, index):
    if parameter[index] != character:
        if index == -1:
            parameter += character
        else:
            parameter = parameter[:index] + character + parameter[index:]
    return parameter


def _check_solved(data):
    if data["solved"] is not True:
        return (
            "{data_type} not yet solved therefore nothing to inspect. "
            "First, run :meth:`~geopyv.{lcdata_type}.{data_type}.solve()` to solve."
        ).format(data_type=data["type"], lcdata_type=data["type"].lower())


def _conversion(parameter, name, new_type, show=True):
    if show:
        return ("`{name}` kwarg converted to a {type}: {value}.").format(
            name=name,
            type=new_type.__name__,
            value=parameter,
        )
    else:
        return ("`{name}` kwarg converted to a {type}.").format(
            name=name,
            type=new_type.__name__,
        )


def _report(msg, error_type):
    if msg and error_type != "Warning":
        log.error(msg)
    elif msg and error_type == "Warning":
        log.warning(msg)
    if error_type == "ValueError":
        raise ValueError(msg)
    elif error_type == "TypeError":
        raise TypeError(msg)



import numpy as np
import geopyv as gp
import matplotlib.pyplot as plt

# Subset test.
# Subset setup.
ref = gp.image.Image("./images/T-Bar/IMG_1062.jpg")
tar = gp.image.Image("./images/T-Bar/IMG_1071.jpg")
template = gp.templates.Circle(50)

# Subset instantiation.
subset = gp.subset.Subset(
    f_img=ref, g_img=tar, f_coord=np.asarray([1000, 1000]), template=template
)

# Subset inspection.
subset.inspect()

# Subset solving.
subset.solve()

# Other subset methods.
subset.convergence()

# Save subset to pyv file.
gp.io.save(object=subset, filename="test")
del subset

# Load subset test.
subset = gp.io.load(filename="test")

# Mesh test.
# Mesh setup.
ref = gp.image.Image("./images/T-Bar/IMG_1062.jpg")
tar = gp.image.Image("./images/T-Bar/IMG_1065.jpg")
template = gp.templates.Circle(50)
boundary = gp.geometry.region.Path(
    nodes=np.asarray(
        [[200.0, 200.0], [200.0, 2700.0], [3900.0, 2700.0], [3900.0, 200.0]]
    ),
    rigid=False,
    hard=False,
    track=True,
)
exclusions = []
exclusions.append(
    gp.geometry.region.Circle(
        centre=np.asarray([1925, 1470]),
        radius=430.0,
        size=100.0,
        rigid=True,
        hard=True,
        track=True,
    )
)
seed = np.asarray([400, 400.0])
alpha = 0.2

# Mesh instantiation.
mesh = gp.mesh.Mesh(
    f_img=ref,
    g_img=tar,
    target_nodes=1000,
    boundary_obj=boundary,
    exclusion_objs=exclusions,
    mesh_order=1,
)

# Mesh inspection.
# mesh.inspect(subset_index=5000)
# Mesh saving : note, prior to solving, a geopyv object cannot be saved.
gp.io.save(object=mesh, filename="mesh")

# Mesh solving.
mesh.solve(
    seed_coord=seed,
    template=template,
    adaptive_iterations=3,
    method="ICGN",
    alpha=alpha,
    tolerance=0.7,
)

# Mesh saving.
gp.io.save(object=mesh, filename="mesh")
del mesh

# Mesh loading.
mesh = gp.io.load(filename="mesh")

# Other mesh methods (plot functionaility).
# The commands are basically standard matplotlib...
mesh.convergence()
mesh.convergence(quantity="norm")
mesh.convergence(quantity="iterations")
mesh.quiver()
mesh.contour(
    quantity="u",
    colorbar=False,
    alpha=0.75,
    levels=np.arange(-5, 6, 1),
    axis=False,
    xlim=((900, 2900)),
    ylim=((500, 2500)),
)
mesh.contour(quantity="C_ZNCC", mesh=True)
mesh.contour(quantity="iterations")
mesh.contour(quantity="R")

# You can also return the fig and ax objects and add customised items to the plot.
fig, ax = mesh.contour(quantity="v", alpha=1.0, levels=np.arange(-5, 6, 1))
ax.plot([0, 2000], [1000, 1000], color="k", linewidth=3.0, zorder=10)
ax.set_xlim((0, 2000))
plt.show()
#
# Let's inspect some clipped subsets.
mesh.inspect(subset_index=0)
mesh.inspect(subset_index=1)
mesh.inspect(subset_index=2)
mesh.inspect(subset_index=3)
mesh.inspect(subset_index=4)

# You can inspect subsets and their convergence
# via the mesh object by passing a subset number.
mesh.inspect(subset_index=0)
mesh.convergence(subset_index=0)

# If you supply a subset index that is out of range you get a ValueError.
# mesh.convergence(subset_index=4000)
# Sequence test.
# Sequence setup.
template = gp.templates.Circle(50)
boundary = gp.geometry.region.Path(
    nodes=np.asarray(
        [[200.0, 200.0], [200.0, 2700.0], [3900.0, 2700.0], [3900.0, 200.0]]
    ),
    rigid=False,
    hard=False,
    track=True,
)
exclusions = []
exclusions.append(
    gp.geometry.region.Circle(
        centre=np.asarray([1925, 1470]),
        radius=430.0,
        size=100.0,
        rigid=True,
        hard=True,
        track=True,
    )
)
seed = np.asarray([400, 400.0])
alpha = 0.2

# Sequence instantiation.
sequence = gp.sequence.Sequence(
    image_dir="images/T-Bar/",
    target_nodes=1000,
    boundary_obj=boundary,
    exclusion_objs=exclusions,
    save_by_reference=True,
    mesh_dir="images/T-Bar/meshes/",
)

# Sequence solving.
sequence.solve(
    seed_coord=seed,
    template=template,
    adaptive_iterations=0,
    method="ICGN",
    alpha=alpha,
    tolerance=0.7,
)

# Sequence saving.
gp.io.save(object=sequence, filename="T_bar_sequence")
del sequence

# Sequence loading.
sequence = gp.io.load(filename="T_bar_sequence")

# Other sequence methods (plot functionality).
sequence.inspect(mesh_index=0)
sequence.inspect(mesh_index=3, subset_index=20)
sequence.convergence(mesh_index=1)
sequence.convergence(mesh_index=2, subset_index=4)
sequence.contour(mesh_index=1, quantity="R", mesh=True)
sequence.quiver(mesh_index=3)

# Particle test.
# Particle setup.
coordinate_0 = np.asarray([1600.0, 1900.0])

# Particle instantiation.
particle = gp.particle.Particle(series=sequence, coordinate_0=coordinate_0)

# Particle solving.
particle.solve()

# Particle saving.
gp.io.save(object=particle, filename="particle")
del particle

# Particle loading.
particle = gp.io.load(filename="particle")

# Other particle methods.
particle.trace(quantity="warps", component=1)

# Field test.
# Field setup.
target_particles = 500

# Field instantiation.
field = gp.field.Field(series=sequence, target_particles=target_particles)

# Field solving.
field.solve()

# Field saving.
gp.io.save(object=field, filename="field")
del field

# Field loading.
field = gp.io.load(filename="field")

# Other field methods.
field.trace(quantity="warps", component=2)
field.inspect()

# Extracting data.
component = 0
particle_index = 4

# Coordinates.
coordinates = field.data["particles"][particle_index]["results"]["coordinates"]
print(coordinates)

# Warps.
# Full.
warps = field.data["particles"][particle_index]["results"]["warps"]  # warp components :
# [u, v, dudx, dudy, dvdx, dvdy, d2udx2, d2udxdy, d2udy2, d2vdx2, d2vdxdy, d2vdy2]
print(warps)
warps = field.data["particles"][particle_index]["results"]["warps"][
    2:5, :2
]  # Time steps 2-5 for components 0 and 1 i.e. displacements.
print(warps)

# e.g. to extract volume progression. Note, area is in pixels.
volume_array = field.data["particles"][particle_index]["results"]["volumes"]
print(volume_array)
volume = field.data["particles"][particle_index]["results"]["volumes"][
    3
]  # At time step 3.
print(volume)



"""

Sequence module for geopyv.

"""
import logging
import numpy as np
import geopyv as gp
from geopyv.object import Object
import re
import glob
import traceback
from alive_progress import alive_bar

log = logging.getLogger(__name__)


class SequenceBase(Object):
    """

    Sequence base class to be used as a mixin. Contains plot functionality.

    """

    def __init__(self):
        super().__init__(object_type="Sequence")
        """

        Sequence base class initialiser.

        """

    def inspect(
        self,
        *,
        mesh_index=None,
        subset_index=None,
        show=True,
        block=True,
        save=None,
    ):
        """Method to show the sequence and associated mesh and subset quality metrics using
        :mod: `~geopyv.plots.inspect_subset` for subsets and
        :mod: `~geopyv.plots.inspect_mesh` for meshes.

        Parameters
        ----------
        mesh_index : int
            Index of the mesh to inspect.
        subset_index : int, optional
            Index of the subset to inspect. If `None', the mesh is inspected instead.
        show : bool, optional
            Control whether the plot is displayed.
        block : bool, optional
            Control whether the plot blocks execution until closed.
        save : str, optional
            Name to use to save plot. Uses default extension of `.png`.

        Returns
        -------
        fig :  `matplotlib.pyplot.figure`
            Figure object.
        ax : `matplotlib.pyplot.axes`
            Axes object.

        .. note::
            * The figure and axes objects can be returned allowing standard
              matplotlib functionality to be used to augment the plot generated.
              See the :ref:`plots tutorial <Plots Tutorial>` for guidance.

        .. seealso::
            :meth:`~geopyv.plots.inspect_subset`
            :meth:`~geopyv.plots.inspect_mesh`
        """

        # Check if solved.
        if self.data["solved"] is not True:
            log.error(
                "Sequence not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.sequence.Sequence.solve()` to solve."
            )
            raise ValueError(
                "Sequence not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.sequence.Sequence.solve()` to solve."
            )

        # Check mesh_index input.
        self._report(
            gp.check._check_type(mesh_index, "mesh_index", [int, type(None)]),
            "TypeError",
        )
        if mesh_index:
            self._report(
                gp.check._check_index(mesh_index, "mesh_index", 0, self.data["meshes"]),
                "IndexError",
            )

        # Load/access selected mesh object.
        mesh_obj = self._load_mesh(mesh_index)

        # Check remaining input.
        self._report(
            gp.check._check_type(subset_index, "subset_index", [int, type(None)]),
            "TypeError",
        )
        if subset_index:
            self._report(
                gp.check._check_index(
                    subset_index, "subset_index", 0, mesh_obj["results"]["subsets"]
                ),
                "IndexError",
            )
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        # Inspect.
        if subset_index is not None:
            subset_data = mesh_obj["results"]["subsets"][subset_index]
            mask = np.asarray(mesh_obj["mask"])
            log.info(
                "Inspecting subset {subset} of mesh {mesh}...".format(
                    subset=subset_index, mesh=mesh_index
                )
            )
            fig, ax = gp.plots.inspect_subset(
                data=subset_data,
                mask=mask,
                show=show,
                block=block,
                save=save,
            )
            return fig, ax
        else:
            log.info("Inspecting mesh {mesh}...".format(mesh=mesh_index))
            fig, ax = gp.plots.inspect_mesh(
                data=mesh_obj,
                show=show,
                block=block,
                save=save,
            )
            return fig, ax

    def convergence(
        self,
        *,
        mesh_index=None,
        subset_index=None,
        quantity=None,
        show=True,
        block=True,
        save=None,
    ):
        """
        Method to plot the rate of convergence using
        :mod: `~geopyv.plots.convergence_subset` for subsets and
        :mod: `~geopyv.plots.convergence_mesh` for meshes.

        Parameters
        ----------
        mesh_index : int
            Index of the mesh to inspect.
        subset_index : int, optional
            Index of the subset to inspect. If `None',
            the convergence plot is for the mesh instead.
        quantity : str, optional
            Selector for histogram convergence property if
            the convergence plot is for mesh. Defaults
            to `C_ZNCC` if left as default None.
        show : bool, optional
            Control whether the plot is displayed.
        block : bool, optional
            Control whether the plot blocks execution until closed.
        save : str, optional
            Name to use to save plot. Uses default extension of `.png`.


        Returns
        -------
        fig :  matplotlib.pyplot.figure
            Figure object.
        ax : `matplotlib.pyplot.axes`
            Axes object.


        .. note::
            * The figure and axes objects can be returned allowing standard
              matplotlib functionality to be used to augment the plot generated.
              See the :ref:`plots tutorial <Plots Tutorial>` for guidance.

        .. warning::
            * Can only be used once the mesh has been solved using the
              :meth:`~geopyv.mesh.Mesh.solve` method.

        .. seealso::
            :mod: `~geopyv.plots.convergence_subset`
            :meth:`~geopyv.plots.convergence_mesh`

        """

        # Check if solved.
        if self.data["solved"] is not True:
            log.error(
                "Sequence not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.sequence.Sequence.solve()` to solve."
            )
            raise ValueError(
                "Sequence not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.sequence.Sequence.solve()` to solve."
            )

        # Check mesh_index input.
        self._report(
            gp.check._check_type(mesh_index, "mesh_index", [int, type(None)]),
            "TypeError",
        )
        if mesh_index:
            self._report(
                gp.check._check_index(mesh_index, "mesh_index", 0, self.data["meshes"]),
                "IndexError",
            )

        # Load/access selected mesh object.
        mesh_obj = self._load_mesh(mesh_index)

        # Check remaining input.
        self._report(
            gp.check._check_type(subset_index, "subset_index", [int, type(None)]),
            "TypeError",
        )
        if subset_index:
            self._report(
                gp.check._check_index(
                    subset_index, "subset_index", 0, mesh_obj["results"]["subsets"]
                ),
                "IndexError",
            )
        self._report(
            gp.check._check_type(quantity, "quantity", [str, type(None)]), "TypeError"
        )
        if quantity:
            self._report(
                gp.check._check_value(
                    quantity, "quantity", ["C_ZNCC", "iterations", "norm"]
                ),
                "ValueError",
            )
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        # Plot convergence.
        if subset_index is not None:
            log.info(
                (
                    "Generating convergence plots for subset {subset} "
                    "of mesh {mesh}..."
                ).format(subset=subset_index, mesh=mesh_index)
            )
            fig, ax = gp.plots.convergence_subset(
                mesh_obj["results"]["subsets"][subset_index],
                show=show,
                block=block,
                save=save,
            )
            return fig, ax
        else:
            if quantity is None:
                quantity = "C_ZNCC"
            elif quantity not in ["C_ZNCC", "iterations", "norm"]:
                log.warning(
                    (
                        "Invalid quantity specified: {quantity}. "
                        "Reverting to C_ZNCC."
                    ).format(quantity=quantity)
                )
                quantity = "C_ZNCC"
            log.info(
                (
                    "Generating {quantity} convergence histogram " "for mesh {mesh}..."
                ).format(quantity=quantity, mesh=mesh_index)
            )
            fig, ax = gp.plots.convergence_mesh(
                data=mesh_obj,
                quantity=quantity,
                show=show,
                block=block,
                save=save,
            )
            return fig, ax

    def contour(
        self,
        *,
        mesh_index=None,
        quantity="C_ZNCC",
        imshow=True,
        colorbar=True,
        ticks=None,
        mesh=False,
        alpha=0.75,
        levels=None,
        axis=True,
        xlim=None,
        ylim=None,
        show=True,
        block=True,
        save=None,
    ):
        """

        Method to plot the contours of a given quantity.

        Parameters
        ----------
        mesh_index : int
            Index of the mesh to inspect.
        quantity : str, optional
            Selector for contour parameter. Must be in:
            [`C_ZNCC`, `u`, `v`, `u_x`, `v_x`,`u_y`,`v_y`, `R`]
            Defaults to `C_ZNCC`.
        imshow : bool, optional
            Control whether the reference image is plotted.
            Defaults to True.
        colorbar : bool, optional
            Control whether the colour bar is plotted.
            Defaults to True.
        ticks : list, optional
            Overwrite default colourbar ticks.
            Defaults to None.
        mesh : bool, optional
            Control whether the mesh is plotted.
            Defaults to False.
        alpha : float, optional
            Control contour opacity. Must be between 0.0-1.0.
            Defaults to 0.75.
        levels : int or array-like, optional
            Control number and position of contour lines.
            Defaults to None.
        axis : bool, optional
            Control whether the axes are plotted.
            Defaults to True.
        xlim : array-like, optional
            Set the plot x-limits (lower_limit,upper_limit).
            Defaults to None.
        ylim : array-like, optional
            Set the plot y-limits (lower_limit,upper_limit).
            Defaults to None.
        show : bool, optional
            Control whether the plot is displayed.
            Defaults to True.
        block : bool, optional
            Control whether the plot blocks execution until closed.
            Defaults to False.
        save : str, optional
            Name to use to save plot. Uses default extension of `.png`.


        Returns
        -------
        fig :  matplotlib.pyplot.figure
            Figure object.
        ax : `matplotlib.pyplot.axes`
            Axes object.


        .. note::
            * The figure and axes objects can be returned allowing standard
              matplotlib functionality to be used to augment the plot generated.
              See the :ref:`plots tutorial <Plots Tutorial>` for guidance.

        .. warning::
            * Can only be used once the sequence has been solved using the
              :meth:`~geopyv.sequence.Sequence.solve` method.

        .. seealso::
            :meth:`~geopyv.plots.contour_sequence`

        """
        # Check if solved.
        if self.data["solved"] is not True:
            log.error(
                "Sequence not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.sequence.Sequence.solve()` to solve."
            )
            raise ValueError(
                "Sequence not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.sequence.Sequence.solve()` to solve."
            )

        # Check input.
        self._report(
            gp.check._check_type(mesh_index, "mesh_index", [int, type(None)]),
            "TypeError",
        )
        if mesh_index:
            self._report(
                gp.check._check_index(mesh_index, "mesh_index", 0, self.data["meshes"]),
                "IndexError",
            )

        self._report(
            gp.check._check_type(quantity, "quantity", [str, type(None)]), "TypeError"
        )
        types = [
            "C_ZNCC",
            "iterations",
            "norm",
            "u",
            "v",
            "u_x",
            "v_x",
            "u_y",
            "v_y",
            "R",
        ]
        if quantity:
            self._report(
                gp.check._check_value(quantity, "quantity", types), "ValueError"
            )
        self._report(gp.check._check_type(imshow, "imshow", [bool]), "TypeError")
        self._report(gp.check._check_type(colorbar, "colorbar", [bool]), "TypeError")
        types = [tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(ticks, "ticks", types), "TypeError")
        self._report(gp.check._check_type(mesh, "mesh", [bool]), "TypeError")
        check = gp.check._check_type(alpha, "alpha", [float])
        if check:
            try:
                alpha = float(alpha)
                self._report(gp.check._conversion(alpha, "alpha", float), "Warning")
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_range(alpha, "alpha", 0.0, 1.0), "ValueError")
        types = [int, tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(levels, "levels", types), "TypeError")
        self._report(gp.check._check_type(axis, "axis", [bool]), "TypeError")
        types = [tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(xlim, "xlim", types), "TypeError")
        if xlim is not None:
            self._report(gp.check._check_dim(xlim, "xlim", 1), "ValueError")
            self._report(gp.check._check_axis(xlim, "xlim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(ylim, "ylim", types), "TypeError")
        if ylim is not None:
            self._report(gp.check._check_dim(ylim, "ylim", 1), "ValueError")
            self._report(gp.check._check_axis(ylim, "ylim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        # Load/access selected mesh object.
        mesh_obj = self._load_mesh(mesh_index)

        log.info(
            "Generating {quantity} contour plot for mesh {mesh}...".format(
                quantity=quantity, mesh=mesh_index
            )
        )
        fig, ax = gp.plots.contour_mesh(
            data=mesh_obj,
            imshow=imshow,
            quantity=quantity,
            colorbar=colorbar,
            ticks=ticks,
            mesh=mesh,
            alpha=alpha,
            levels=levels,
            axis=axis,
            xlim=xlim,
            ylim=ylim,
            show=show,
            block=block,
            save=save,
        )
        return fig, ax

    def quiver(
        self,
        *,
        mesh_index=None,
        scale=1.0,
        imshow=True,
        mesh=False,
        axis=True,
        xlim=None,
        ylim=None,
        show=True,
        block=True,
        save=None,
    ):
        """

        Method to plot a quiver plot of the displacements.

        Parameters
        ----------
        mesh_index : int
            Index of the mesh to inspect.
        scale : float, optional
            Control size of quiver arrows.
        imshow : bool, optional
            Control whether the reference image is plotted.
            Defaults to True.
        mesh : bool, optional
            Control whether the mesh is plotted.
            Defaults to False.
        axis : bool, optional
            Control whether the axes are plotted.
            Defaults to True.
        xlim : array-like, optional
            Set the plot x-limits (lower_limit,upper_limit).
            Defaults to None.
        ylim : array-like, optional
            Set the plot y-limits (lower_limit,upper_limit).
            Defaults to None.
        show : bool, optional
            Control whether the plot is displayed.
            Defaults to True.
        block : bool, optional
            Control whether the plot blocks execution until closed.
            Defaults to False.
        save : str, optional
            Name to use to save plot. Uses default extension of `.png`.

        Returns
        -------
        fig :  matplotlib.pyplot.figure
            Figure object.
        ax : `matplotlib.pyplot.axes`
            Axes object.


        .. note::
            * The figure and axes objects can be returned allowing standard
              matplotlib functionality to be used to augment the plot generated.
              See the :ref:`plots tutorial <Plots Tutorial>` for guidance.

        .. warning::
            * Can only be used once the mesh has been solved using the
              :meth:`~geopyv.mesh.Mesh.solve` method.

        .. seealso::
            :meth:`~geopyv.plots.quiver_mesh`
        """
        # Check if solved.
        if self.data["solved"] is not True:
            log.error(
                "Sequence not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.sequence.Sequence.solve()` to solve."
            )
            raise ValueError(
                "Sequence not yet solved therefore nothing to inspect. "
                "First, run :meth:`~geopyv.sequence.Sequence.solve()` to solve."
            )

        # Check input.
        self._report(
            gp.check._check_type(mesh_index, "mesh_index", [int, type(None)]),
            "TypeError",
        )
        if mesh_index:
            self._report(
                gp.check._check_index(mesh_index, "mesh_index", 0, self.data["meshes"]),
                "IndexError",
            )
        check = gp.check._check_type(scale, "scale", [float])
        if check:
            try:
                scale = float(scale)
                self._report(gp.check._conversion(scale, "scale", float), "Warning")
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_range(scale, "scale", 0.0), "ValueError")
        self._report(gp.check._check_type(imshow, "imshow", [bool]), "TypeError")
        self._report(gp.check._check_type(mesh, "mesh", [bool]), "TypeError")
        self._report(gp.check._check_type(axis, "axis", [bool]), "TypeError")
        types = [tuple, list, np.ndarray, type(None)]
        self._report(gp.check._check_type(xlim, "xlim", types), "TypeError")
        if xlim is not None:
            self._report(gp.check._check_dim(xlim, "xlim", 1), "ValueError")
            self._report(gp.check._check_axis(xlim, "xlim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(ylim, "ylim", types), "TypeError")
        if ylim is not None:
            self._report(gp.check._check_dim(ylim, "ylim", 1), "ValueError")
            self._report(gp.check._check_axis(ylim, "ylim", 0, [2]), "ValueError")
        self._report(gp.check._check_type(show, "show", [bool]), "TypeError")
        self._report(gp.check._check_type(block, "block", [bool]), "TypeError")
        self._report(gp.check._check_type(save, "save", [str, type(None)]), "TypeError")

        # Load/access selected mesh object.
        mesh_obj = self._load_mesh(mesh_index)

        # Plot quiver.
        log.info("Generating quiver plot for mesh {mesh}...".format(mesh=mesh_index))
        fig, ax = gp.plots.quiver_mesh(
            data=mesh_obj,
            scale=scale,
            imshow=imshow,
            mesh=mesh,
            axis=axis,
            xlim=xlim,
            ylim=ylim,
            show=show,
            block=block,
            save=save,
        )
        return fig, ax

    def _load_mesh(self, mesh_index, verbose=True):
        if self.data["file_settings"]["save_by_reference"]:
            return gp.io.load(
                filename=self.data["file_settings"]["mesh_dir"]
                + self.data["meshes"][mesh_index],
                verbose=verbose,
            ).data
        else:
            return self.data["meshes"][mesh_index]

    def _report(self, msg, error_type):
        if msg and error_type != "Warning":
            log.error(msg)
        elif msg and error_type == "Warning":
            log.warning(msg)
            return True
        if error_type == "ValueError" and msg:
            raise ValueError(msg)
        elif error_type == "TypeError" and msg:
            raise TypeError(msg)
        elif error_type == "IndexError" and msg:
            raise IndexError(msg)


class Sequence(SequenceBase):
    def __init__(
        self,
        *,
        image_dir=".",
        common_name="",
        file_format=".jpg",
        target_nodes=1000,
        boundary_obj=None,
        exclusion_objs=[],
        size_lower_bound=1.0,
        size_upper_bound=1000.0,
        save_by_reference=False,
        mesh_dir=".",
    ):
        """Initialisation of geopyv sequence object.

        Parameters
        ----------
        image_dir : str, optional
            Directory of images. Defaults to current working directory.
        file_format : str, optional
            Image file type. Options are ".jpg", ".png" or ".bmp". Defaults to .jpg.
        target_nodes : int
            Target node number. Defaults to 1000.
        boundary : `numpy.ndarray` (N,2)
            Array of coordinates to define the mesh boundary.
            Must be specified in clockwise or anti-clockwise order.
        exclusions: list, optional
            List of `numpy.ndarray` to define the mesh exclusions.
        size_lower_bound : int, optional
            Lower bound on element size. Defaults to a value of 1.
        upper_lower_bound : int, optional
            Lower bound on element size. Defaults to a value of 1000.
        save_by_reference : bool, optional
            False - sequence object stores meshes.
            True - sequence object stores mesh references.

        Attributes
        ----------
        data : dict
            Data object containing all settings and results.
            See the data structure :ref:`here <sequence_data_structure>`.
        solved : bool
            Boolean to indicate if the sequence has been solved.
        """

        # Set initialised boolean.
        self.initialised = False

        # Check types.
        self._report(gp.check._check_type(image_dir, "image_dir", [str]), "TypeError")
        if self._report(gp.check._check_path(image_dir, "image_dir"), "Warning"):
            image_dir = gp.io._get_image_dir()
        image_dir = gp.check._check_character(image_dir, "/", -1)
        self._report(
            gp.check._check_type(file_format, "file_format", [str]), "TypeError"
        )
        file_format = gp.check._check_character(file_format, ".", 0)
        self._report(
            gp.check._check_value(
                file_format,
                "file_format",
                [".jpg", ".png", ".bmp", ".JPG", ".PNG", ".BMP"],
            ),
            "ValueError",
        )
        check = gp.check._check_type(target_nodes, "target_nodes", [int])
        if check:
            try:
                target_nodes = int(target_nodes)
                self._report(
                    gp.check._conversion(target_nodes, "target_nodes", int), "Warning"
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(target_nodes, "target_nodes", 1), "ValueError"
        )
        self._report(
            gp.check._check_type(
                boundary_obj,
                "boundary_obj",
                [gp.geometry.region.Circle, gp.geometry.region.Path],
            ),
            "TypeError",
        )
        check = gp.check._check_type(exclusion_objs, "exclusion_objs", [list])
        if check:
            try:
                exclusion_objs = list(exclusion_objs)
                self._report(
                    gp.check._conversion(exclusion_objs, "exclusion_objs", list, False),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        for exclusion_obj in exclusion_objs:
            self._report(
                gp.check._check_type(
                    exclusion_obj,
                    "exclusion_obj",
                    [gp.geometry.region.Circle, gp.geometry.region.Path],
                ),
                "TypeError",
            )
        self._report(
            gp.check._check_type(size_lower_bound, "size_lower_bound", [int, float]),
            "TypeError",
        )
        self._report(
            gp.check._check_type(size_upper_bound, "size_upper_bound", [int, float]),
            "TypeError",
        )
        self._report(
            gp.check._check_comp(
                size_lower_bound,
                "size_lower_bound",
                size_upper_bound,
                "size_upper_bound",
            ),
            "ValueError",
        )
        self._report(
            gp.check._check_type(save_by_reference, "save_by_reference", [bool]),
            "TypeError",
        )
        self._report(gp.check._check_type(mesh_dir, "mesh_dir", [str]), "TypeError")
        if self._report(gp.check._check_path(mesh_dir, "mesh_dir"), "Warning"):
            mesh_dir = gp.io._get_mesh_dir()
        mesh_dir = gp.check._check_character(mesh_dir, "/", -1)

        # Store variables.
        self._image_dir = image_dir
        self._file_format = file_format
        self._common_name = common_name
        try:
            _images = glob.glob(
                self._common_name + "*" + self._file_format, root_dir=self._image_dir
            )
            if np.shape(_images)[0] < 2:
                log.error("Insufficient number of images in specified folder.")
                raise FileExistsError(
                    "Insufficient number of images in specified folder."
                )
            _image_indices_unordered = [int(re.findall(r"\d+", x)[-1]) for x in _images]
            _image_indices_arguments = np.argsort(_image_indices_unordered)
            self._images = [_images[index] for index in _image_indices_arguments]
            self._image_indices = np.sort(_image_indices_unordered)

        except Exception:
            log.error(
                (
                    "No files found under: \n{address}"
                    "Please check file names and directories."
                ).format(
                    address=self._image_dir
                    + self._common_name
                    + "*"
                    + self._file_format
                )
            )
            print(traceback.format_exc())
            return
        self._target_nodes = target_nodes
        self._boundary_obj = boundary_obj
        self._exclusion_objs = exclusion_objs
        self._size_lower_bound = size_lower_bound
        self._size_upper_bound = size_upper_bound
        self._save_by_reference = save_by_reference
        self._mesh_dir = mesh_dir
        if self._save_by_reference:
            meshes = []
        else:
            meshes = np.empty(np.shape(self._image_indices)[0] - 1, dtype=dict)
        self.solved = False
        self._unsolvable = False

        # Data.
        file_settings = {
            "image_dir": self._image_dir,
            "images": self._images,
            "file_format": self._file_format,
            "image_indices": self._image_indices,
            "save_by_reference": self._save_by_reference,
            "mesh_dir": self._mesh_dir,
        }
        mesh_settings = {
            "target_nodes": self._target_nodes,
            "boundary_obj": self._boundary_obj,
            "exclusion_objs": self._exclusion_objs,
            "size_lower_bound": self._size_lower_bound,
            "size_upper_bound": self._size_upper_bound,
        }
        self.data = {
            "type": "Sequence",
            "solved": self.solved,
            "unsolvable": self._unsolvable,
            "meshes": meshes,
            "mesh_settings": mesh_settings,
            "file_settings": file_settings,
        }

        self._initialised = True

    def solve(
        self,
        *,
        seed_coord=None,
        seed_warp=None,
        template=None,
        max_norm=1e-5,
        max_iterations=50,
        adaptive_iterations=0,
        corrective_iterations=0,
        method="ICGN",
        mesh_order=2,
        subset_order=1,
        tolerance=0.75,
        seed_tolerance=0.9,
        alpha=0.5,
        guide=False,
        sequential=False,
    ):
        """
        Method to solve for the sequence.

        Parameters
        ----------
        seed_coord : numpy.ndarray (2,)
            Seed coordinate for reliability-guided mesh solving.
        template : gp.template.Template object, optional
            subset template. Defaults to Circle(50).
        max_norm : float, optional
            Exit criterion for norm of increment in warp function.
            Defaults to value of
            :math:`1 cdot 10^{-3}`.
        max_iterations : int, optional
            Exit criterion for number of Gauss-Newton iterations.
            Defaults to value of 15.
        mesh_order : int, optional
            Mesh element order. Options are 1 and 2. Defaults to 2.
        subset_order : int, optional
            Warp function order. Options are 1 and 2. Defaults to 1.
        tolerance: float, optional
            Correlation coefficient tolerance. Defaults to a value of 0.7.
        seed_tolerance: float, optional
            Correlation coefficient tolerance for the seed subset.
            Defaults to a value of 0.9.
        method : str
            Solution method. Options are FAGN and ICGN.
            Default is ICGN since it is faster.
        adaptive_iterations : int, optional
            Number of mesh adaptivity iterations to perform.
            Defaults to a value of 0.
        alpha : float, optional
            Mesh adaptivity control parameter.
            Defaults to a value of 0.5.

        Returns
        -------

        solved : bool
            Boolean to indicate if the subset instance has been solved.
        """
        # Check if solved.
        if self.data["solved"] is True:
            log.error("Sequence has already been solved. Cannot be solved again.")
            return

        # Check inputs.
        if self._report(
            gp.check._check_type(seed_coord, "seed_coord", [np.ndarray]), "Warning"
        ):
            seed_coord = gp.gui.selectors.coordinate.CoordinateSelector()
        elif self._report(gp.check._check_dim(seed_coord, "seed_coord", 1), "Warning"):
            seed_coord = gp.gui.selectors.coordinate.CoordinateSelector()
        elif self._report(
            gp.check._check_axis(seed_coord, "seed_coord", 0, [2]), "Warning"
        ):
            seed_coord = gp.gui.selectors.coordinate.CoordinateSelector()
        if template is None:
            template = gp.templates.Circle(50)
        types = [gp.templates.Circle, gp.templates.Square]
        self._report(gp.check._check_type(template, "template", types), "TypeError")
        check = gp.check._check_type(max_norm, "max_norm", [float])
        if check:
            try:
                max_norm = float(max_norm)
                self._report(
                    gp.check._conversion(max_norm, "max_norm", float), "Warning"
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_range(max_norm, "max_norm", 1e-20), "ValueError")
        check = gp.check._check_type(max_iterations, "max_iterations", [int])
        if check:
            try:
                max_iterations = int(max_iterations)
                self._report(
                    gp.check._conversion(max_iterations, "max_iterations", int),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(max_iterations, "max_iterations", 0.5), "ValueError"
        )
        check = gp.check._check_type(subset_order, "subset_order", [int])
        if check:
            try:
                subset_order = int(subset_order)
                self._report(
                    gp.check._conversion(subset_order, "subset_order", int), "Warning"
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_value(subset_order, "subset_order", [1, 2]), "ValueError"
        )
        check = gp.check._check_type(adaptive_iterations, "adaptive_iterations", [int])
        if check:
            try:
                adaptive_iterations = int(adaptive_iterations)
                self._report(
                    gp.check._conversion(
                        adaptive_iterations, "adaptive_iterations", int
                    ),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(adaptive_iterations, "adaptive_iterations", 0),
            "ValueError",
        )
        check = gp.check._check_type(tolerance, "tolerance", [float])
        if check:
            try:
                tolerance = float(tolerance)
                self._report(
                    gp.check._conversion(tolerance, "tolerance", float), "Warning"
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(tolerance, "tolerance", 0.0, 1.0), "ValueError"
        )
        check = gp.check._check_type(seed_tolerance, "seed_tolerance", [float])
        if check:
            try:
                seed_tolerance = float(seed_tolerance)
                self._report(
                    gp.check._conversion(seed_tolerance, "seed_tolerance", float),
                    "Warning",
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_range(seed_tolerance, "seed_tolerance", 0.0, 1.0),
            "ValueError",
        )
        self._report(
            gp.check._check_comp(
                tolerance,
                "tolerance",
                seed_tolerance,
                "seed_tolerance",
            ),
            "ValueError",
        )
        self._report(gp.check._check_type(method, "method", [str]), "TypeError")
        self._report(
            gp.check._check_value(method, "method", ["FAGN", "ICGN"]), "ValueError"
        )
        check = gp.check._check_type(mesh_order, "mesh_order", [int])
        if check:
            try:
                mesh_order = int(mesh_order)
                self._report(
                    gp.check._conversion(mesh_order, "mesh_order", int), "Warning"
                )
            except Exception:
                self._report(check, "TypeError")
        self._report(
            gp.check._check_value(mesh_order, "mesh_order", [1, 2]), "ValueError"
        )
        check = gp.check._check_type(alpha, "alpha", [float])
        if check:
            try:
                alpha = float(alpha)
                self._report(gp.check._conversion(alpha, "alpha", float), "Warning")
            except Exception:
                self._report(check, "TypeError")
        self._report(gp.check._check_range(alpha, "alpha", 0.0, 1.0), "ValueError")
        if seed_warp is not None:
            check = gp.check._check_type(seed_warp, "seed_warp", [np.ndarray])
            if check:
                try:
                    seed_warp = np.asarray(seed_warp)
                    self._report(
                        gp.check._conversion(seed_warp, "seed_warp", np.ndarray),
                        "Warning",
                    )
                except Exception:
                    self._report(check, "TypeError")
            self._report(gp.check._check_dim(seed_warp, "seed_warp", 1), "ValueError")
            self._report(
                gp.check._check_axis(seed_warp, "seed_warp", 0, [6 * subset_order]),
                "ValueError",
            )
        else:
            seed_warp = np.zeros(12)

        # Store variables.
        self._seed_coord = seed_coord
        self._template = template
        self._max_iterations = max_iterations
        self._max_norm = max_norm
        self._adaptive_iterations = adaptive_iterations
        self._corrective_iterations = corrective_iterations
        self._method = method
        self._subset_order = subset_order
        self._mesh_order = mesh_order
        self._tolerance = tolerance
        self._seed_tolerance = seed_tolerance
        self._alpha = alpha
        self._guide = guide
        self._sequential = sequential
        self._seed_warp = np.zeros(6 * self._subset_order)

        # Solve.
        _f_index = 0
        _g_index = 1
        _seed_coord_t = seed_coord
        _seed_displacement = np.zeros(2)
        _f_img = gp.image.Image(self._image_dir + self._images[_f_index])
        _g_img = gp.image.Image(self._image_dir + self._images[_g_index])
        while _g_index < len(self._image_indices):
            log.info(
                "Solving for image pair {}-{}.".format(
                    self._image_indices[_f_index],
                    self._image_indices[_g_index],
                )
            )
            mesh = gp.mesh.Mesh(
                f_img=_f_img,
                g_img=_g_img,
                target_nodes=self._target_nodes,
                boundary_obj=self._boundary_obj,
                exclusion_objs=self._exclusion_objs,
                size_lower_bound=self._size_lower_bound,
                size_upper_bound=self._size_upper_bound,
                mesh_order=self._mesh_order,
            )  # Initialise mesh object.
            mesh.solve(
                seed_coord=self._seed_coord,
                seed_warp=self._seed_warp,
                template=self._template,
                max_iterations=self._max_iterations,
                max_norm=self._max_norm,
                adaptive_iterations=self._adaptive_iterations,
                corrective_iterations=self._corrective_iterations,
                method=self._method,
                subset_order=self._subset_order,
                tolerance=self._tolerance,
                seed_tolerance=self._seed_tolerance,
                alpha=self._alpha,
            )  # Solve mesh.
            if mesh.solved:
                if self._save_by_reference:
                    gp.io.save(
                        object=mesh,
                        filename=self._mesh_dir
                        + "mesh_"
                        + str(self._image_indices[_f_index])
                        + "_"
                        + str(self._image_indices[_g_index]),
                    )
                    self.data["meshes"].append(
                        "mesh_"
                        + str(self._image_indices[_f_index])
                        + "_"
                        + str(self._image_indices[_g_index])
                    )
                else:
                    self.data["meshes"][_g_index - 1] = mesh.data
                if self._guide:
                    seed = gp.particle.Particle(series=mesh, coordinate_0=_seed_coord_t)
                    seed.solve()
                    _seed_displacement += seed.data["results"]["warps"][1, :2]
                    self._seed_warp[
                        : 6 * min(self._mesh_order, self._subset_order)
                    ] = seed.data["results"]["warps"][
                        1, : 6 * min(self._mesh_order, self._subset_order)
                    ]
                _g_index += 1  # Iterate the target image index.
                del _g_img
                if _g_index != len(self._image_indices):
                    _g_img = gp.image.Image(self._image_dir + self._images[_g_index])
                else:
                    self.solved = True
                if self._sequential:
                    _f_index = _g_index - 1
                    del _f_img
                    _f_img = gp.image.Image(self._image_dir + self._images[_f_index])
                    _seed_coord_t += _seed_displacement
                    _seed_displacement = np.zeros(2)
            elif _f_index + 1 < _g_index:
                _f_index = _g_index - 1
                del _f_img
                _f_img = gp.image.Image(self._image_dir + self._images[_f_index])
                _seed_coord_t += _seed_displacement
                _seed_displacement = np.zeros(2)
            else:
                log.error(
                    "Mesh for consecutive image pair {a}-{b} is unsolvable. "
                    "Sequence curtailed.".format(
                        a=self._image_indices[_f_index],
                        b=self._image_indices[_g_index],
                    )
                )
                self.data["meshes"] = np.asarray(self.data["meshes"])
                self._unsolvable = True
                del mesh
                return self.solved
            del mesh
        del _f_img

        # Pack data.
        self.data["solved"] = self.solved
        self.data["unsolvable"] = self._unsolvable
        self.data["meshes"] = np.asarray(self.data["meshes"])
        return self.solved

    def load(self):
        try:
            _meshes = glob.glob("*.pyv", root_dir=self._mesh_dir)
            _mesh_tars = [int(re.findall(r"\d+", x)[-1]) for x in _meshes]
            _mesh_indices_arguments = np.argsort(_mesh_tars)
            self._meshes = np.asarray(
                [_meshes[index] for index in _mesh_indices_arguments], dtype=object
            )
        except Exception:
            log.error(
                "Issues encountered recognising mesh file names. "
                "Please refer to the documentation for naming guidance."
            )
        self._images = self._images[: np.shape(self._meshes)[0] + 1]
        self.data["solved"] = True
        self.data["unsolvable"] = False
        self.data["meshes"] = self._meshes
        self.data["file_settings"]["images"]


class SequenceResults(SequenceBase):
    """
    SequenceResults class for geopyv.

    Parameters
    ----------
    data : dict
        geopyv data dict from Sequence object.


    Attributes
    ----------
    data : dict
        geopyv data dict from Sequence object.

    """

    def __init__(self, data):
        """Initialisation of geopyv SequenceResults class."""
        self.data = data

    def load(self, ref=True):
        """Load all meshes in sequence object mesh directory."""
        try:
            _meshes = glob.glob(
                "*.pyv", root_dir=self.data["file_settings"]["mesh_dir"]
            )
            _mesh_tars = [int(re.findall(r"\d+", x)[-1]) for x in _meshes]
            _mesh_indices_arguments = np.argsort(_mesh_tars)
            self._meshes = np.asarray(
                [_meshes[index] for index in _mesh_indices_arguments], dtype=object
            )
        except Exception:
            log.error(
                "Issues encountered recognising mesh file names. "
                "Please refer to the documentation for naming guidance."
            )
        self._images = self.data["file_settings"]["images"][
            : np.shape(self._meshes)[0] + 1
        ]
        self.data["solved"] = True
        self.data["unsolvable"] = False
        self.data["file_settings"]["images"] = self._images
        if ref is not True:
            with alive_bar(
                np.shape(self._meshes)[0],
                dual_line=True,
                bar="blocks",
                title="Loading meshes...",
            ) as bar:
                for i in range(np.shape(self._meshes)[0]):
                    self._meshes[i] = self._load_mesh(mesh_index=i, verbose=False)
                    bar()
            self.data["file_settings"]["save_by_reference"] = False
        self.data["meshes"] = self._meshes
